# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Data plugin represeting an executable code to be wrapped and called through a `CalcJob` plugin."""
import os
import warnings

from aiida.common import exceptions
from aiida.common.warnings import AiidaDeprecationWarning
from .data import Data

__all__ = ('Code',)


class Code(Data):
    """
    A code entity.
    It can either be 'local', or 'remote'.

    * Local code: it is a collection of files/dirs (added using the add_path() method), where one \
    file is flagged as executable (using the set_local_executable() method).

    * Remote code: it is a pair (remotecomputer, remotepath_of_executable) set using the \
    set_remote_computer_exec() method.

    For both codes, one can set some code to be executed right before or right after
    the execution of the code, using the set_preexec_code() and set_postexec_code()
    methods (e.g., the set_preexec_code() can be used to load specific modules required
    for the code to be run).
    """

    # pylint: disable=too-many-public-methods

    def __init__(self, remote_computer_exec=None, local_executable=None, input_plugin_name=None, files=None, **kwargs):
        super().__init__(**kwargs)

        if remote_computer_exec and local_executable:
            raise ValueError('cannot set `remote_computer_exec` and `local_executable` at the same time')

        if remote_computer_exec:
            self.set_remote_computer_exec(remote_computer_exec)

        if local_executable:
            self.set_local_executable(local_executable)

        if input_plugin_name:
            self.set_input_plugin_name(input_plugin_name)

        if files:
            self.set_files(files)

    HIDDEN_KEY = 'hidden'

    def hide(self):
        """
        Hide the code (prevents from showing it in the verdi code list)
        """
        self.set_extra(self.HIDDEN_KEY, True)

    def reveal(self):
        """
        Reveal the code (allows to show it in the verdi code list)
        By default, it is revealed
        """
        self.set_extra(self.HIDDEN_KEY, False)

    @property
    def hidden(self):
        """
        Determines whether the Code is hidden or not
        """
        return self.get_extra(self.HIDDEN_KEY, False)

    def set_files(self, files):
        """
        Given a list of filenames (or a single filename string),
        add it to the path (all at level zero, i.e. without folders).
        Therefore, be careful for files with the same name!

        :todo: decide whether to check if the Code must be a local executable
             to be able to call this function.
        """

        if isinstance(files, str):
            files = [files]

        for filename in files:
            if os.path.isfile(filename):
                with open(filename, 'rb') as handle:
                    self.put_object_from_filelike(handle, os.path.split(filename)[1], 'wb', encoding=None)

    def __str__(self):
        local_str = 'Local' if self.is_local() else 'Remote'
        computer_str = self.computer.label
        return f"{local_str} code '{self.label}' on {computer_str}, pk: {self.pk}, uuid: {self.uuid}"

    def get_computer_name(self):
        """Get label of this code's computer.

        .. deprecated:: 1.4.0
            Will be removed in `v2.0.0`, use the `self.get_computer_label()` method instead.
        """
        return self.get_computer_label()

    def get_computer_label(self):
        """Get label of this code's computer."""
        return 'repository' if self.is_local() else self.computer.label

    @property
    def full_label(self):
        """Get full label of this code.

        Returns label of the form <code-label>@<computer-name>.
        """
        return f'{self.label}@{self.get_computer_label()}'

    @property
    def label(self):
        """Return the node label.

        :return: the label
        """
        return super().label

    @label.setter
    def label(self, value):
        """Set the label.

        :param value: the new value to set
        """
        if '@' in str(value):
            msg = "Code labels must not contain the '@' symbol"
            raise exceptions.InputValidationError(msg)

        super(Code, self.__class__).label.fset(self, value)  # pylint: disable=no-member

    def relabel(self, new_label, raise_error=True):
        """Relabel this code.

        :param new_label: new code label
        :param raise_error: Set to False in order to return a list of errors
            instead of raising them.

        .. deprecated:: 1.2.0
            Will remove raise_error in `v2.0.0`. Use `try/except` instead.
        """
        # pylint: disable=unused-argument
        suffix = f'@{self.computer.label}'
        if new_label.endswith(suffix):
            new_label = new_label[:-len(suffix)]

        self.label = new_label

    def get_description(self):
        """Return a string description of this Code instance.

        :return: string description of this Code instance
        """
        return f'{self.description}'

    @classmethod
    def get_code_helper(cls, label, machinename=None):
        """
        :param label: the code label identifying the code to load
        :param machinename: the machine name where code is setup

        :raise aiida.common.NotExistent: if no code identified by the given string is found
        :raise aiida.common.MultipleObjectsError: if the string cannot identify uniquely
            a code
        """
        from aiida.common.exceptions import NotExistent, MultipleObjectsError
        from aiida.orm.querybuilder import QueryBuilder
        from aiida.orm.computers import Computer

        query = QueryBuilder()
        query.append(cls, filters={'label': label}, project='*', tag='code')
        if machinename:
            query.append(Computer, filters={'name': machinename}, with_node='code')

        if query.count() == 0:
            raise NotExistent(f"'{label}' is not a valid code name.")
        elif query.count() > 1:
            codes = query.all(flat=True)
            retstr = f"There are multiple codes with label '{label}', having IDs: "
            retstr += f"{', '.join(sorted([str(c.pk) for c in codes]))}.\n"
            retstr += ('Relabel them (using their ID), or refer to them with their ID.')
            raise MultipleObjectsError(retstr)
        else:
            return query.first()[0]

    @classmethod
    def get(cls, pk=None, label=None, machinename=None):
        """
        Get a Computer object with given identifier string, that can either be
        the numeric ID (pk), or the label (and computername) (if unique).

        :param pk: the numeric ID (pk) for code
        :param label: the code label identifying the code to load
        :param machinename: the machine name where code is setup

        :raise aiida.common.NotExistent: if no code identified by the given string is found
        :raise aiida.common.MultipleObjectsError: if the string cannot identify uniquely a code
        :raise aiida.common.InputValidationError: if neither a pk nor a label was passed in
        """
        # pylint: disable=arguments-differ
        from aiida.orm.utils import load_code

        # first check if code pk is provided
        if pk:
            code_int = int(pk)
            try:
                return load_code(pk=code_int)
            except exceptions.NotExistent:
                raise ValueError(f'{pk} is not valid code pk')
            except exceptions.MultipleObjectsError:
                raise exceptions.MultipleObjectsError(f"More than one code in the DB with pk='{pk}'!")

        # check if label (and machinename) is provided
        elif label is not None:
            return cls.get_code_helper(label, machinename)

        else:
            raise exceptions.InputValidationError('Pass either pk or code label (and machinename)')

    @classmethod
    def get_from_string(cls, code_string):
        """
        Get a Computer object with given identifier string in the format
        label@machinename. See the note below for details on the string
        detection algorithm.

        .. note:: the (leftmost) '@' symbol is always used to split code
            and computername. Therefore do not use
            '@' in the code name if you want to use this function
            ('@' in the computer name are instead valid).

        :param code_string: the code string identifying the code to load

        :raise aiida.common.NotExistent: if no code identified by the given string is found
        :raise aiida.common.MultipleObjectsError: if the string cannot identify uniquely
            a code
        :raise aiida.common.InputValidationError: if code_string is not of string type

        """
        from aiida.common.exceptions import NotExistent, MultipleObjectsError, InputValidationError

        try:
            label, _, machinename = code_string.partition('@')
        except AttributeError:
            raise InputValidationError('the provided code_string is not of valid string type')

        try:
            return cls.get_code_helper(label, machinename)
        except NotExistent:
            raise NotExistent(f'{code_string} could not be resolved to a valid code label')
        except MultipleObjectsError:
            raise MultipleObjectsError(f'{code_string} could not be uniquely resolved')

    @classmethod
    def list_for_plugin(cls, plugin, labels=True):
        """
        Return a list of valid code strings for a given plugin.

        :param plugin: The string of the plugin.
        :param labels: if True, return a list of code names, otherwise
          return the code PKs (integers).
        :return: a list of string, with the code names if labels is True,
          otherwise a list of integers with the code PKs.
        """
        from aiida.orm.querybuilder import QueryBuilder
        query = QueryBuilder()
        query.append(cls, filters={'attributes.input_plugin': {'==': plugin}})
        valid_codes = query.all(flat=True)

        if labels:
            return [c.label for c in valid_codes]

        return [c.pk for c in valid_codes]

    def _validate(self):
        super()._validate()

        if self.is_local() is None:
            raise exceptions.ValidationError('You did not set whether the code is local or remote')

        if self.is_local():
            if not self.get_local_executable():
                raise exceptions.ValidationError(
                    'You have to set which file is the local executable '
                    'using the set_exec_filename() method'
                )
            if self.get_local_executable() not in self.list_object_names():
                raise exceptions.ValidationError(
                    "The local executable '{}' is not in the list of "
                    'files of this code'.format(self.get_local_executable())
                )
        else:
            if self.list_object_names():
                raise exceptions.ValidationError('The code is remote but it has files inside')
            if not self.get_remote_computer():
                raise exceptions.ValidationError('You did not specify a remote computer')
            if not self.get_remote_exec_path():
                raise exceptions.ValidationError('You did not specify a remote executable')

    def set_prepend_text(self, code):
        """
        Pass a string of code that will be put in the scheduler script before the
        execution of the code.
        """
        self.set_attribute('prepend_text', str(code))

    def get_prepend_text(self):
        """
        Return the code that will be put in the scheduler script before the
        execution, or an empty string if no pre-exec code was defined.
        """
        return self.get_attribute('prepend_text', '')

    def set_input_plugin_name(self, input_plugin):
        """
        Set the name of the default input plugin, to be used for the automatic
        generation of a new calculation.
        """
        if input_plugin is None:
            self.set_attribute('input_plugin', None)
        else:
            self.set_attribute('input_plugin', str(input_plugin))

    def get_input_plugin_name(self):
        """
        Return the name of the default input plugin (or None if no input plugin
        was set.
        """
        return self.get_attribute('input_plugin', None)

    def set_append_text(self, code):
        """
        Pass a string of code that will be put in the scheduler script after the
        execution of the code.
        """
        self.set_attribute('append_text', str(code))

    def get_append_text(self):
        """
        Return the postexec_code, or an empty string if no post-exec code was defined.
        """
        return self.get_attribute('append_text', '')

    def set_local_executable(self, exec_name):
        """
        Set the filename of the local executable.
        Implicitly set the code as local.
        """
        self._set_local()
        self.set_attribute('local_executable', exec_name)

    def get_local_executable(self):
        return self.get_attribute('local_executable', '')

    def set_remote_computer_exec(self, remote_computer_exec):
        """
        Set the code as remote, and pass the computer on which it resides
        and the absolute path on that computer.

        :param remote_computer_exec: a tuple (computer, remote_exec_path), where computer is a aiida.orm.Computer and
            remote_exec_path is the absolute path of the main executable on remote computer.
        """
        from aiida import orm
        from aiida.common.lang import type_check

        if (not isinstance(remote_computer_exec, (list, tuple)) or len(remote_computer_exec) != 2):
            raise ValueError(
                'remote_computer_exec must be a list or tuple '
                'of length 2, with machine and executable '
                'name'
            )

        computer, remote_exec_path = tuple(remote_computer_exec)

        if not os.path.isabs(remote_exec_path):
            raise ValueError('exec_path must be an absolute path (on the remote machine)')

        type_check(computer, orm.Computer)

        self._set_remote()
        self.computer = computer
        self.set_attribute('remote_exec_path', remote_exec_path)

    def get_remote_exec_path(self):
        if self.is_local():
            raise ValueError('The code is local')
        return self.get_attribute('remote_exec_path', '')

    def get_remote_computer(self):
        if self.is_local():
            raise ValueError('The code is local')

        return self.computer

    def _set_local(self):
        """
        Set the code as a 'local' code, meaning that all the files belonging to the code
        will be copied to the cluster, and the file set with set_exec_filename will be
        run.

        It also deletes the flags related to the local case (if any)
        """
        self.set_attribute('is_local', True)
        self.computer = None
        try:
            self.delete_attribute('remote_exec_path')
        except AttributeError:
            pass

    def _set_remote(self):
        """
        Set the code as a 'remote' code, meaning that the code itself has no files attached,
        but only a location on a remote computer (with an absolute path of the executable on
        the remote computer).

        It also deletes the flags related to the local case (if any)
        """
        self.set_attribute('is_local', False)
        try:
            self.delete_attribute('local_executable')
        except AttributeError:
            pass

    def is_local(self):
        """
        Return True if the code is 'local', False if it is 'remote' (see also documentation
        of the set_local and set_remote functions).
        """
        return self.get_attribute('is_local', None)

    def can_run_on(self, computer):
        """
        Return True if this code can run on the given computer, False otherwise.

        Local codes can run on any machine; remote codes can run only on the machine
        on which they reside.

        TODO: add filters to mask the remote machines on which a local code can run.
        """
        from aiida import orm
        from aiida.common.lang import type_check

        if self.is_local():
            return True

        type_check(computer, orm.Computer)
        return computer.id == self.get_remote_computer().id

    def get_execname(self):
        """
        Return the executable string to be put in the script.
        For local codes, it is ./LOCAL_EXECUTABLE_NAME
        For remote codes, it is the absolute path to the executable.
        """
        if self.is_local():
            return f'./{self.get_local_executable()}'

        return self.get_remote_exec_path()

    def get_builder(self):
        """Create and return a new `ProcessBuilder` for the `CalcJob` class of the plugin configured for this code.

        The configured calculation plugin class is defined by the `get_input_plugin_name` method.

        .. note:: it also sets the ``builder.code`` value.

        :return: a `ProcessBuilder` instance with the `code` input already populated with ourselves
        :raise aiida.common.EntryPointError: if the specified plugin does not exist.
        :raise ValueError: if no default plugin was specified.
        """
        from aiida.plugins import CalculationFactory

        plugin_name = self.get_input_plugin_name()

        if plugin_name is None:
            raise ValueError('no default calculation input plugin specified for this code')

        try:
            process_class = CalculationFactory(plugin_name)
        except exceptions.EntryPointError:
            raise exceptions.EntryPointError(f'the calculation entry point `{plugin_name}` could not be loaded')

        builder = process_class.get_builder()
        builder.code = self

        return builder

    def get_full_text_info(self, verbose=False):
        """Return a list of lists with a human-readable detailed information on this code.

        .. deprecated:: 1.4.0
            Will be removed in `v2.0.0`.

        :return: list of lists where each entry consists of two elements: a key and a value
        """
        warnings.warn('this property is deprecated', AiidaDeprecationWarning)  # pylint: disable=no-member
        from aiida.repository import FileType

        result = []
        result.append(['PK', self.pk])
        result.append(['UUID', self.uuid])
        result.append(['Label', self.label])
        result.append(['Description', self.description])
        result.append(['Default plugin', self.get_input_plugin_name()])

        if verbose:
            result.append(['Calculations', len(self.get_outgoing().all())])

        if self.is_local():
            result.append(['Type', 'local'])
            result.append(['Exec name', self.get_execname()])
            result.append(['List of files/folders:', ''])
            for obj in self.list_objects():
                if obj.file_type == FileType.DIRECTORY:
                    result.append(['directory', obj.name])
                else:
                    result.append(['file', obj.name])
        else:
            result.append(['Type', 'remote'])
            result.append(['Remote machine', self.get_remote_computer().label])
            result.append(['Remote absolute path', self.get_remote_exec_path()])

        if self.get_prepend_text().strip():
            result.append(['Prepend text', ''])
            for line in self.get_prepend_text().split('\n'):
                result.append(['', line])
        else:
            result.append(['Prepend text', 'No prepend text'])

        if self.get_append_text().strip():
            result.append(['Append text', ''])
            for line in self.get_append_text().split('\n'):
                result.append(['', line])
        else:
            result.append(['Append text', 'No append text'])

        return result



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi run` command."""
import contextlib
import os
import functools
import sys
import warnings

import click

from aiida.cmdline.commands.cmd_verdi import verdi
from aiida.cmdline.params.options.multivalue import MultipleValueOption
from aiida.cmdline.utils import decorators, echo
from aiida.common.warnings import AiidaDeprecationWarning


@contextlib.contextmanager
def update_environment(argv):
    """Context manager that temporarily replaces `sys.argv` with `argv` and adds current working dir to the path."""
    try:
        # Store a copy of the current path and argv as a backup variable so it can be restored later
        _path = sys.path[:]
        _argv = sys.argv[:]

        # Add the current working directory to the path, such that local modules can be imported
        sys.path.append(os.getcwd())
        sys.argv = argv[:]
        yield
    finally:
        # Restore old parameters when exiting from the context manager
        sys.argv = _argv
        sys.path = _path


def validate_entrypoint_string(ctx, param, value):  # pylint: disable=unused-argument,invalid-name
    """Validate that `value` is a valid entrypoint string."""
    from aiida.orm import autogroup

    try:
        autogroup.Autogroup.validate(value)
    except Exception as exc:
        raise click.BadParameter(f'{str(exc)} ({value})')

    return value


@verdi.command('run', context_settings=dict(ignore_unknown_options=True,))
@click.argument('scriptname', type=click.STRING)
@click.argument('varargs', nargs=-1, type=click.UNPROCESSED)
@click.option('--auto-group', is_flag=True, help='Enables the autogrouping')
@click.option(
    '-l',
    '--auto-group-label-prefix',
    type=click.STRING,
    required=False,
    help='Specify the prefix of the label of the auto group (numbers might be automatically '
    'appended to generate unique names per run).'
)
@click.option(
    '-n',
    '--group-name',
    type=click.STRING,
    required=False,
    help='Specify the name of the auto group [DEPRECATED, USE --auto-group-label-prefix instead]. '
    'This also enables auto-grouping.'
)
@click.option(
    '-e',
    '--exclude',
    cls=MultipleValueOption,
    default=None,
    help='Exclude these classes from auto grouping (use full entrypoint strings).',
    callback=functools.partial(validate_entrypoint_string)
)
@click.option(
    '-i',
    '--include',
    cls=MultipleValueOption,
    default=None,
    help='Include these classes from auto grouping  (use full entrypoint strings or "all").',
    callback=validate_entrypoint_string
)
@decorators.with_dbenv()
def run(scriptname, varargs, auto_group, auto_group_label_prefix, group_name, exclude, include):
    # pylint: disable=too-many-arguments,exec-used
    """Execute scripts with preloaded AiiDA environment."""
    from aiida.cmdline.utils.shell import DEFAULT_MODULES_LIST
    from aiida.orm import autogroup

    # Prepare the environment for the script to be run
    globals_dict = {
        '__builtins__': globals()['__builtins__'],
        '__name__': '__main__',
        '__file__': scriptname,
        '__doc__': None,
        '__package__': None
    }

    # Dynamically load modules (the same of verdi shell) - but in globals_dict, not in the current environment
    for app_mod, model_name, alias in DEFAULT_MODULES_LIST:
        globals_dict[f'{alias}'] = getattr(__import__(app_mod, {}, {}, model_name), model_name)

    if group_name:
        warnings.warn('--group-name is deprecated, use `--auto-group-label-prefix` instead', AiidaDeprecationWarning)  # pylint: disable=no-member
        if auto_group_label_prefix:
            raise click.BadParameter(
                'You cannot specify both --group-name and --auto-group-label-prefix; '
                'use --auto-group-label-prefix only'
            )
        auto_group_label_prefix = group_name
        # To have the old behavior, with auto-group enabled.
        auto_group = True

    if auto_group:
        aiida_verdilib_autogroup = autogroup.Autogroup()
        # Set the ``group_label_prefix`` if defined, otherwise a default prefix will be used
        if auto_group_label_prefix is not None:
            aiida_verdilib_autogroup.set_group_label_prefix(auto_group_label_prefix)
        aiida_verdilib_autogroup.set_exclude(exclude)
        aiida_verdilib_autogroup.set_include(include)

        # Note: this is also set in the exec environment! This is the intended behavior
        autogroup.CURRENT_AUTOGROUP = aiida_verdilib_autogroup

    # Initialize the variable here, otherwise we get UnboundLocalError in the finally clause if it fails to open
    handle = None

    try:
        # Here we use a standard open and not open, as exec will later fail if passed a unicode type string.
        handle = open(scriptname, 'r')
    except IOError:
        echo.echo_critical(f"Unable to load file '{scriptname}'")
    else:
        try:
            # Must add also argv[0]
            argv = [scriptname] + list(varargs)
            with update_environment(argv=argv):
                # Compile the script for execution and pass it to exec with the globals_dict
                exec(compile(handle.read(), scriptname, 'exec', dont_inherit=True), globals_dict)  # yapf: disable # pylint: disable=exec-used
        except SystemExit:  # pylint: disable=try-except-raise
            # Script called sys.exit()
            # Re-raise the exception to have the error code properly returned at the end
            raise
    finally:
        autogroup.current_autogroup = None
        if handle:
            handle.close()



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name
"""Various utils that should be used during migrations and migrations tests because the AiiDA ORM cannot be used."""

import datetime
import errno
import os
import re

import numpy

from aiida.common import json

ISOFORMAT_DATETIME_REGEX = re.compile(r'^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}\.\d+(\+\d{2}:\d{2})?$')


def ensure_repository_folder_created(uuid):
    """Make sure that the repository sub folder for the node with the given UUID exists or create it.

    :param uuid: UUID of the node
    """
    dirpath = get_node_repository_sub_folder(uuid)

    try:
        os.makedirs(dirpath)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise


def put_object_from_string(uuid, name, content):
    """Write a file with the given content in the repository sub folder of the given node.

    :param uuid: UUID of the node
    :param name: name to use for the file
    :param content: the content to write to the file
    """
    ensure_repository_folder_created(uuid)
    filepath = os.path.join(get_node_repository_sub_folder(uuid), name)

    with open(filepath, 'w', encoding='utf-8') as handle:
        handle.write(content)


def get_object_from_repository(uuid, name):
    """Return the content of a file with the given name in the repository sub folder of the given node.

    :param uuid: UUID of the node
    :param name: name to use for the file
    """
    filepath = os.path.join(get_node_repository_sub_folder(uuid), name)

    with open(filepath) as handle:
        return handle.read()


def get_node_repository_sub_folder(uuid):
    """Return the absolute path to the sub folder `path` within the repository of the node with the given UUID.

    :param uuid: UUID of the node
    :return: absolute path to node repository folder, i.e `/some/path/repository/node/12/ab/c123134-a123/path`
    """
    from aiida.manage.configuration import get_profile

    uuid = str(uuid)

    repo_dirpath = os.path.join(get_profile().repository_path, 'repository')
    node_dirpath = os.path.join(repo_dirpath, 'node', uuid[:2], uuid[2:4], uuid[4:], 'path')

    return node_dirpath


def get_numpy_array_absolute_path(uuid, name):
    """Return the absolute path of a numpy array with the given name in the repository of the node with the given uuid.

    :param uuid: the UUID of the node
    :param name: the name of the numpy array
    :return: the absolute path of the numpy array file
    """
    return os.path.join(get_node_repository_sub_folder(uuid), f'{name}.npy')


def store_numpy_array_in_repository(uuid, name, array):
    """Store a numpy array in the repository folder of a node.

    :param uuid: the node UUID
    :param name: the name under which to store the array
    :param array: the numpy array to store
    """
    ensure_repository_folder_created(uuid)
    filepath = get_numpy_array_absolute_path(uuid, name)

    with open(filepath, 'wb') as handle:
        numpy.save(handle, array)


def delete_numpy_array_from_repository(uuid, name):
    """Delete the numpy array with a given name from the repository corresponding to a node with a given uuid.

    :param uuid: the UUID of the node
    :param name: the name of the numpy array
    """
    filepath = get_numpy_array_absolute_path(uuid, name)

    try:
        os.remove(filepath)
    except (IOError, OSError):
        pass


def load_numpy_array_from_repository(uuid, name):
    """Load and return a numpy array from the repository folder of a node.

    :param uuid: the node UUID
    :param name: the name under which to store the array
    :return: the numpy array
    """
    filepath = get_numpy_array_absolute_path(uuid, name)
    return numpy.load(filepath)


def recursive_datetime_to_isoformat(value):
    """Convert all datetime objects in the given value to string representations in ISO format.

    :param value: a mapping, sequence or single value optionally containing datetime objects
    """
    if isinstance(value, list):
        return [recursive_datetime_to_isoformat(_) for _ in value]

    if isinstance(value, dict):
        return dict((key, recursive_datetime_to_isoformat(val)) for key, val in value.items())

    if isinstance(value, datetime.datetime):
        return value.isoformat()

    return value


def dumps_json(dictionary):
    """Transforms all datetime object into isoformat and then returns the JSON."""
    return json.dumps(recursive_datetime_to_isoformat(dictionary))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
Defines an rst directive to auto-document AiiDA processes.
"""
from collections.abc import Mapping, Iterable
import inspect

from docutils import nodes
from docutils.core import publish_doctree
from docutils.parsers.rst import directives
from sphinx import addnodes
from sphinx.ext.autodoc import ClassDocumenter
from sphinx.util.docutils import SphinxDirective
from sphinxcontrib.details.directive import details, summary

from plumpy.ports import OutputPort

from aiida.common.utils import get_object_from_string
from aiida.engine import Process
from aiida.engine.processes.ports import InputPort, PortNamespace
from aiida.manage.configuration import load_profile


def setup_extension(app):
    app.add_directive_to_domain('py', AiidaProcessDocumenter.directivetype, AiidaProcessDirective)
    # app.add_autodocumenter(AiidaProcessDocumenter)


class AiidaProcessDocumenter(ClassDocumenter):
    """Sphinx Documenter class for AiiDA Processes."""
    directivetype = 'aiida-process'
    objtype = 'process'
    priority = 10

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        return inspect.isclass(member) and issubclass(member, Process)


class AiidaProcessDirective(SphinxDirective):
    """
    Directive to auto-document AiiDA processes.
    """
    required_arguments = 1
    optional_arguments = 0
    final_argument_whitespace = True

    HIDE_UNSTORED_INPUTS_FLAG = 'hide-nondb-inputs'
    EXPAND_NAMESPACES_FLAG = 'expand-namespaces'
    option_spec = {
        'module': directives.unchanged_required,
        HIDE_UNSTORED_INPUTS_FLAG: directives.flag,
        EXPAND_NAMESPACES_FLAG: directives.flag
    }
    signature = 'Process'
    annotation = 'process'

    has_content = True

    def run(self):
        self.initialize()
        return self.build_node_tree()

    def initialize(self):
        """Set internal attributes of the class.

        Includes importing the process class.
        """
        # pylint: disable=attribute-defined-outside-init
        load_profile()

        self.class_name = self.arguments[0].split('(')[0]
        self.module_name = self.options['module']
        self.process_name = f'{self.module_name}.{self.class_name}'
        self.process = get_object_from_string(self.process_name)

        try:
            self.process_spec = self.process.spec()
        except Exception as exc:
            raise RuntimeError(f"Error while building the spec for process '{self.process_name}': '{exc!r}.'") from exc

    def build_node_tree(self):
        """Returns the docutils node tree."""
        process_node = addnodes.desc(desctype='class', domain='py', noindex=False, objtype='class')
        process_node += self.build_signature()
        process_node += self.build_content()
        return [process_node]

    def build_signature(self):
        """Returns the signature of the process."""
        signature = addnodes.desc_signature(first=False, fullname=self.signature)
        signature += addnodes.desc_annotation(text=self.annotation)
        signature += addnodes.desc_addname(text=f'{self.module_name}.')
        signature += addnodes.desc_name(text=self.class_name)
        return signature

    def build_content(self):
        """
        Returns the main content (docstring, inputs, outputs) of the documentation.
        """
        content = addnodes.desc_content()

        content += nodes.paragraph(text=self.process.__doc__)

        content += self.build_doctree(
            title='Inputs:',
            port_namespace=self.process_spec.inputs,
        )
        content += self.build_doctree(title='Outputs:', port_namespace=self.process_spec.outputs)

        if hasattr(self.process_spec, 'get_outline'):
            outline = self.process_spec.get_outline()
            if outline is not None:
                content += self.build_outline_doctree(outline=outline)
        return content

    def build_doctree(self, title, port_namespace):
        """
        Returns a doctree for a given port namespace, including a title.
        """
        paragraph = nodes.paragraph()
        paragraph += nodes.strong(text=title)
        namespace_doctree = self.build_portnamespace_doctree(port_namespace)
        if namespace_doctree:
            paragraph += namespace_doctree
        else:
            paragraph += nodes.paragraph(text='None defined.')

        return paragraph

    def build_portnamespace_doctree(self, port_namespace):
        """Builds the doctree for a port namespace."""
        if not port_namespace:
            return None
        result = nodes.bullet_list(bullet='*')
        for name, port in sorted(port_namespace.items()):
            item = nodes.list_item()
            if _is_non_db(port) and self.HIDE_UNSTORED_INPUTS_FLAG in self.options:
                continue
            if isinstance(port, (InputPort, OutputPort)):
                item.extend(self.build_port_content(name, port))
            elif isinstance(port, PortNamespace):
                item += addnodes.literal_strong(text=name)
                item += nodes.Text(', ')
                item += nodes.emphasis(text='Namespace')
                if port.help is not None:
                    item += nodes.Text(' -- ')
                    item.extend(publish_doctree(port.help)[0].children)
                sub_doctree = self.build_portnamespace_doctree(port)
                if sub_doctree:
                    sub_item = details(opened=self.EXPAND_NAMESPACES_FLAG in self.options)
                    sub_item += summary(text='Namespace Ports')
                    sub_item += sub_doctree
                    item += sub_item
            else:
                raise NotImplementedError
            result += item
        return result

    def build_port_content(self, name, port):
        """
        Build the content that describes a single port.
        """
        res = []
        res.append(addnodes.literal_strong(text=name))
        res.append(nodes.Text(', '))
        res.append(nodes.emphasis(text=self.format_valid_types(port.valid_type)))
        res.append(nodes.Text(', '))
        res.append(nodes.Text('required' if port.required else 'optional'))
        if _is_non_db(port):
            res.append(nodes.Text(', '))
            res.append(nodes.emphasis(text='non_db'))
        if port.help:
            res.append(nodes.Text(' -- '))
            # publish_doctree returns <document: <paragraph...>>.
            # Here we only want the content (children) of the paragraph.
            res.extend(publish_doctree(port.help)[0].children)
        return res

    @staticmethod
    def format_valid_types(valid_type):
        """Format valid types."""
        try:
            return valid_type.__name__
        except AttributeError:
            try:
                return f"({', '.join(v.__name__ for v in valid_type)})"
            except (AttributeError, TypeError):
                return str(valid_type)

    def build_outline_doctree(self, outline):
        """Build the doctree for a spec outline."""
        paragraph = nodes.paragraph()
        paragraph += nodes.strong(text='Outline:')
        outline_str = '\n'.join(self.build_outline_lines(outline.get_description(), indent=0))
        paragraph += nodes.literal_block(outline_str, outline_str)
        return paragraph

    def build_outline_lines(self, outline, indent):
        """Return a list of lines which describe the process outline."""
        indent_str = ' ' * indent
        res = []
        if isinstance(outline, str):
            res.append(indent_str + outline)
        else:
            if isinstance(outline, Mapping):
                for key, outline_part in outline.items():
                    res.append(indent_str + key)
                    res.extend(self.build_outline_lines(outline_part, indent=indent + 4))
            else:
                assert isinstance(outline, Iterable)
                for outline_part in outline:
                    res.extend(self.build_outline_lines(outline_part, indent=indent))
        return res


def _is_non_db(port):
    return getattr(port, 'non_db', False)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Definition of modules that are to be automatically loaded for `verdi shell`."""

DEFAULT_MODULES_LIST = [
    ('aiida.common.links', 'LinkType', 'LinkType'),
    ('aiida.orm', 'Node', 'Node'),
    ('aiida.orm', 'ProcessNode', 'ProcessNode'),
    ('aiida.orm', 'CalculationNode', 'CalculationNode'),
    ('aiida.orm', 'CalcJobNode', 'CalcJobNode'),
    ('aiida.orm', 'CalcFunctionNode', 'CalcFunctionNode'),
    ('aiida.orm', 'WorkflowNode', 'WorkflowNode'),
    ('aiida.orm', 'WorkChainNode', 'WorkChainNode'),
    ('aiida.orm', 'WorkFunctionNode', 'WorkFunctionNode'),
    ('aiida.orm', 'Data', 'Data'),
    ('aiida.orm', 'Bool', 'Bool'),
    ('aiida.orm', 'Float', 'Float'),
    ('aiida.orm', 'Int', 'Int'),
    ('aiida.orm', 'Str', 'Str'),
    ('aiida.orm', 'List', 'List'),
    ('aiida.orm', 'Dict', 'Dict'),
    ('aiida.orm', 'Code', 'Code'),
    ('aiida.orm', 'Computer', 'Computer'),
    ('aiida.orm', 'Group', 'Group'),
    ('aiida.orm', 'QueryBuilder', 'QueryBuilder'),
    ('aiida.orm', 'User', 'User'),
    ('aiida.orm', 'load_code', 'load_code'),
    ('aiida.orm', 'load_computer', 'load_computer'),
    ('aiida.orm', 'load_group', 'load_group'),
    ('aiida.orm', 'load_node', 'load_node'),
    ('aiida.plugins', 'CalculationFactory', 'CalculationFactory'),
    ('aiida.plugins', 'DataFactory', 'DataFactory'),
    ('aiida.plugins', 'WorkflowFactory', 'WorkflowFactory'),
]


def ipython():
    """Start any version of IPython"""
    for ipy_version in (_ipython, _ipython_pre_100, _ipython_pre_011):
        try:
            ipy_version()
        except ImportError:
            pass
        else:
            return

    raise ImportError('No IPython available')


def bpython():
    """Start a bpython shell."""
    import bpython as bpy_shell  # pylint: disable=import-error

    user_ns = get_start_namespace()
    if user_ns:
        bpy_shell.embed(user_ns)
    else:
        bpy_shell.embed()


AVAILABLE_SHELLS = {'ipython': ipython, 'bpython': bpython}


def run_shell(interface=None):
    """Start the chosen external shell."""

    available_shells = [AVAILABLE_SHELLS[interface]] if interface else AVAILABLE_SHELLS.values()

    # Try the specified or the available shells one by one until you
    # find one that is available. If the wanted shell is not available
    # then an ImportError is raised leading the the loading of a generic
    # shell.
    for curr_shell in available_shells:
        try:
            curr_shell()
            return
        except ImportError:
            pass
    raise ImportError


def get_start_namespace():
    """Load all default and custom modules"""
    from aiida.manage.configuration import get_config

    user_ns = {}

    config = get_config()

    # Load default modules
    for app_mod, model_name, alias in DEFAULT_MODULES_LIST:
        user_ns[alias] = getattr(__import__(app_mod, {}, {}, model_name), model_name)

    verdi_shell_auto_import = config.get_option('verdi.shell.auto_import', config.current_profile.name).split(':')

    # Load custom modules
    modules_list = [(str(e[0]), str(e[2])) for e in [p.rpartition('.') for p in verdi_shell_auto_import] if e[1] == '.']

    for app_mod, model_name in modules_list:
        try:
            user_ns[model_name] = getattr(__import__(app_mod, {}, {}, model_name), model_name)
        except AttributeError:
            # If the module does not exist, we ignore it
            pass

    return user_ns


def _ipython_pre_011():
    """Start IPython pre-0.11"""
    from IPython.Shell import IPShell  # pylint: disable=import-error,no-name-in-module

    user_ns = get_start_namespace()
    if user_ns:
        ipy_shell = IPShell(argv=[], user_ns=user_ns)
    else:
        ipy_shell = IPShell(argv=[])
    ipy_shell.mainloop()


def _ipython_pre_100():
    """Start IPython pre-1.0.0"""
    from IPython.frontend.terminal.ipapp import TerminalIPythonApp  # pylint: disable=import-error,no-name-in-module

    app = TerminalIPythonApp.instance()
    app.initialize(argv=[])
    user_ns = get_start_namespace()
    if user_ns:
        app.shell.user_ns.update(user_ns)
    app.start()


def _ipython():
    """Start IPython >= 1.0"""
    from IPython import start_ipython  # pylint: disable=import-error,no-name-in-module

    user_ns = get_start_namespace()
    if user_ns:
        start_ipython(argv=[], user_ns=user_ns)
    else:
        start_ipython(argv=[])



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
""""Implementation of `DbImporter` for the Materials Project database."""
import datetime
import os
import requests

from pymatgen.ext.matproj import MPRester

from aiida.tools.dbimporters.baseclasses import CifEntry, DbImporter, DbSearchResults


class MaterialsProjectImporter(DbImporter):
    """
    Database importer for the Materials Project.
    """

    _properties = 'structure'
    _supported_keywords = None

    def __init__(self, **kwargs):
        """
        Instantiate the MaterialsProjectImporter by setting up the Materials API (MAPI)
        connection details.
        """
        self.setup_db(**kwargs)

    def setup_db(self, **kwargs):
        """
        Setup the required parameters to the REST API
        """
        try:
            api_key = kwargs['api_key']
        except KeyError:
            try:
                api_key = os.environ['PMG_MAPI_KEY']
            except KeyError as exception:
                raise KeyError(
                    'API key not supplied and PMG_MAPI_KEY environment variable not set. Either pass it when '
                    'initializing the class, or set the environment variable PMG_MAPI_KEY to your API key.'
                ) from exception
            api_key = None

        self._api_key = api_key
        self._verify_api_key()
        self._mpr = MPRester(self._api_key)

    def _verify_api_key(self):
        """
        Verify the supplied API key by issuing a request to Materials Project.
        """
        response = requests.get(
            'https://www.materialsproject.org/rest/v1/api_check', headers={'X-API-KEY': self._api_key}
        )
        response_content = response.json()  # a dict
        if 'error' in response_content:
            raise RuntimeError(response_content['error'])
        if not response_content['valid_response']:
            raise RuntimeError('Materials Project did not give a valid response for the API key check.')
        if not response_content['api_key_valid']:
            raise RuntimeError('Your API key for Materials Project is not valid.')

    @property
    def api_key(self):
        """
        Return the API key configured for the importer
        """
        return self._api_key

    @property
    def properties(self):
        """
        Return the properties that will be queried
        """
        return self._properties

    def get_supported_keywords(self):
        """
        Returns the list of all supported query keywords

        :return: list of strings
        """
        return self._supported_keywords

    def query(self, **kwargs):
        """
        Query the database with a given dictionary of query parameters for a given properties

        :param query: a dictionary with the query parameters
        :param properties: the properties to query
        """
        try:
            query = kwargs['query']
        except AttributeError:
            raise AttributeError(
                'Make sure the supplied dictionary has `query` as a key. This '
                'should contain a dictionary with the right query needed.'
            )
        try:
            properties = kwargs['properties']
        except AttributeError:
            raise AttributeError('Make sure the supplied dictionary has `properties` as a key.')

        if not isinstance(query, dict):
            raise TypeError('The query argument should be a dictionary')

        if properties is None:
            properties = self._properties

        if properties != 'structure':
            raise ValueError(f'Unsupported properties: {properties}')

        results = []
        properties_list = ['material_id', 'cif']
        for entry in self._find(query, properties_list):
            results.append(entry)
        search_results = MaterialsProjectSearchResults(results, return_class=MaterialsProjectCifEntry)

        return search_results

    def _find(self, query, properties):
        """
        Query the database with a given dictionary of query parameters

        :param query: a dictionary with the query parameters
        """
        for entry in self._mpr.query(criteria=query, properties=properties):
            yield entry


class MaterialsProjectCifEntry(CifEntry):  # pylint: disable=abstract-method
    """
    A Materials Project entry class which extends the DbEntry class with a CifEntry class.

    """

    _license = 'Materials Project'

    def __init__(self, url, **kwargs):
        """
        The DbSearchResults base class instantiates a new DbEntry by explicitly passing the url
        of the entry as an argument. In this case it is the same as the 'uri' value that is
        already contained in the source dictionary so we just copy it
        """
        cif = kwargs.pop('cif', None)
        kwargs['uri'] = url
        super().__init__(**kwargs)

        if cif is not None:
            self.cif = cif


class MaterialsProjectSearchResults(DbSearchResults):  # pylint: disable=abstract-method
    """
    A collection of MaterialsProjectEntry query result entries.

    :param results: query result entry dictionary
    :return_class: the class associated with each
    """

    _db_name = 'Materials Project'
    _db_uri = 'https://materialsproject.org'
    _material_base_url = 'https://materialsproject.org/materials/'
    _version = str(datetime.datetime.now())

    def __init__(self, results, return_class):
        self._return_class = return_class
        super().__init__(results)

    def _get_source_dict(self, result_dict):
        """
        Return the source information dictionary of an Materials Project query result entry

        :param result_dict: query result entry dictionary
        """
        source_dict = {
            'db_name': self._db_name,
            'db_uri': self._db_uri,
            'id': result_dict['material_id'],
            'uri': self._get_url(result_dict),
            'version': self._version,
        }

        if 'cif' in result_dict:
            source_dict['cif'] = result_dict['cif']

        return source_dict

    def _get_url(self, result_dict):
        """
        Return the permanent URI of the result entry

        :param result_dict: query result entry dictionary
        """
        return self._material_base_url + result_dict['material_id']



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
.. py:module::conditional
    :synopsis: Tools for options which are required only if a a set of
    conditions on the context are fulfilled
"""

import click


class ConditionalOption(click.Option):
    """
    This cli option takes an additional callable parameter and uses that
    to determine wether a MissingParam should be raised if the option is
    not given on the cli.

    The callable takes the context as an argument and can look up any
    amount of other parameter values etc.

    :param required_fn: callable(ctx) -> True | False, returns True
        if the parameter is required to have a value.
        This is typically used when the condition depends on other
        parameters specified on the command line.
    """

    def __init__(self, param_decls=None, required_fn=None, **kwargs):

        # note default behaviour for required: False
        self.required_fn = required_fn

        # Required_fn overrides 'required', if defined
        if required_fn is not None:
            # There is a required_fn
            self.required = False  # So it does not show up as 'required'

        super().__init__(param_decls=param_decls, **kwargs)

    def full_process_value(self, ctx, value):
        try:
            value = super().full_process_value(ctx, value)
            if self.required_fn and self.value_is_missing(value):
                if self.is_required(ctx):
                    raise click.MissingParameter(ctx=ctx, param=self)
        except click.MissingParameter:
            if self.is_required(ctx):
                raise
        return value

    def is_required(self, ctx):
        """runs the given check on the context to determine requiredness"""

        if self.required_fn:
            return self.required_fn(ctx)

        return self.required



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Module for defintion of base `Data` sub class for numeric based data types."""

from .base import to_aiida_type, BaseType

__all__ = ('NumericType',)


def _left_operator(func):
    """Function decorator to treat a method as the left operator."""

    def inner(self, other):
        """Decorator wrapper."""
        left = self.value
        if isinstance(other, NumericType):
            right = other.value
        else:
            right = other
        return to_aiida_type(func(left, right))

    return inner


def _right_operator(func):
    """Function decorator to treat a method as the right operator."""

    def inner(self, other):
        """Decorator wrapper."""
        assert not isinstance(other, NumericType)
        return to_aiida_type(func(self.value, other))

    return inner


class NumericType(BaseType):
    """Sub class of Data to store numbers, overloading common operators (``+``, ``*``, ...)."""

    @_left_operator
    def __add__(self, other):
        return self + other

    @_right_operator
    def __radd__(self, other):
        return other + self

    @_left_operator
    def __sub__(self, other):
        return self - other

    @_right_operator
    def __rsub__(self, other):
        return other - self

    @_left_operator
    def __mul__(self, other):
        return self * other

    @_right_operator
    def __rmul__(self, other):
        return other * self

    @_left_operator
    def __div__(self, other):
        return self / other

    @_right_operator
    def __rdiv__(self, other):
        return other / self

    @_left_operator
    def __truediv__(self, other):
        return self / other

    @_right_operator
    def __rtruediv__(self, other):
        return other / self

    @_left_operator
    def __floordiv__(self, other):
        return self // other

    @_right_operator
    def __rfloordiv__(self, other):
        return other // self

    @_left_operator
    def __pow__(self, power):
        return self**power

    @_left_operator
    def __lt__(self, other):
        return self < other

    @_left_operator
    def __le__(self, other):
        return self <= other

    @_left_operator
    def __gt__(self, other):
        return self > other

    @_left_operator
    def __ge__(self, other):
        return self >= other

    @_left_operator
    def __mod__(self, other):
        return self % other

    @_right_operator
    def __rmod__(self, other):
        return other % self

    def __float__(self):
        return float(self.value)

    def __int__(self):
        return int(self.value)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Archive reader classes."""
from abc import ABC, abstractmethod
import json
import os
from pathlib import Path
import tarfile
from types import TracebackType
from typing import Any, Callable, cast, Dict, Iterable, Iterator, List, Optional, Set, Tuple, Type
import zipfile

from distutils.version import StrictVersion
from archive_path import TarPath, ZipPath, read_file_in_tar, read_file_in_zip

from aiida.common.log import AIIDA_LOGGER
from aiida.common.exceptions import InvalidOperation
from aiida.common.folders import Folder, SandboxFolder
from aiida.tools.importexport.common.config import EXPORT_VERSION, ExportFileFormat, NODES_EXPORT_SUBFOLDER
from aiida.tools.importexport.common.exceptions import (CorruptArchive, IncompatibleArchiveVersionError)
from aiida.tools.importexport.archive.common import (ArchiveMetadata, null_callback)
from aiida.tools.importexport.common.config import NODE_ENTITY_NAME, GROUP_ENTITY_NAME
from aiida.tools.importexport.common.utils import export_shard_uuid

__all__ = (
    'ArchiveReaderAbstract',
    'ARCHIVE_READER_LOGGER',
    'ReaderJsonBase',
    'ReaderJsonFolder',
    'ReaderJsonTar',
    'ReaderJsonZip',
    'get_reader',
)

ARCHIVE_READER_LOGGER = AIIDA_LOGGER.getChild('archive.reader')


def get_reader(file_format: str) -> Type['ArchiveReaderAbstract']:
    """Return the available writer classes."""
    readers = {
        ExportFileFormat.ZIP: ReaderJsonZip,
        ExportFileFormat.TAR_GZIPPED: ReaderJsonTar,
        'folder': ReaderJsonFolder,
    }

    if file_format not in readers:
        raise ValueError(
            f'Can only read in the formats: {tuple(readers.keys())}, please specify one for "file_format".'
        )

    return cast(Type[ArchiveReaderAbstract], readers[file_format])


class ArchiveReaderAbstract(ABC):
    """An abstract interface for AiiDA archive readers.

    An ``ArchiveReader`` implementation is intended to be used with a context::

        with ArchiveReader(filename) as reader:
            reader.entity_count('Node')

    """

    def __init__(self, filename: str, **kwargs: Any):
        """An archive reader

        :param filename: the filename (possibly including the absolute path)
            of the file to import.

        """
        # pylint: disable=unused-argument
        self._filename = filename
        self._in_context = False

    @property
    def filename(self) -> str:
        """Return the name of the file that is being read from."""
        return self._filename

    @property
    @abstractmethod
    def file_format_verbose(self) -> str:
        """The file format name."""

    @property
    @abstractmethod
    def compatible_export_version(self) -> str:
        """Return the export version that this reader is compatible with."""

    def __enter__(self) -> 'ArchiveReaderAbstract':
        self._in_context = True
        return self

    def __exit__(
        self, exctype: Optional[Type[BaseException]], excinst: Optional[BaseException], exctb: Optional[TracebackType]
    ):
        self._in_context = False

    def assert_within_context(self):
        """Assert that the method is called within a context.

        :raises: `~aiida.common.exceptions.InvalidOperation`: if not called within a context
        """
        if not self._in_context:
            raise InvalidOperation('the ArchiveReader method should be used within a context')

    @property
    @abstractmethod
    def export_version(self) -> str:
        """Return the export version.

        :raises `~aiida.tools.importexport.common.exceptions.CorruptArchive`: If the version cannot be retrieved.
        """
        # this should be able to be returned independent of any metadata validation

    def check_version(self):
        """Check the version compatibility of the archive.

        :raises: `~aiida.tools.importexport.common.exceptions.IncompatibleArchiveVersionError`:
            If the version is not compatible

        """
        file_version = StrictVersion(self.export_version)
        expected_version = StrictVersion(self.compatible_export_version)

        try:
            if file_version != expected_version:
                msg = f'Archive file version is {file_version}, can read only version {expected_version}'
                if file_version < expected_version:
                    msg += "\nUse 'verdi export migrate' to update this archive file."
                else:
                    msg += '\nUpdate your AiiDA version in order to import this file.'

                raise IncompatibleArchiveVersionError(msg)
        except AttributeError:
            msg = (
                f'Archive file version is {self.export_version}, '
                f'can read only version {self.compatible_export_version}'
            )
            raise IncompatibleArchiveVersionError(msg)

    @property
    @abstractmethod
    def metadata(self) -> ArchiveMetadata:
        """Return the full (validated) archive metadata."""

    @property
    def entity_names(self) -> List[str]:
        """Return list of all entity names."""
        return list(self.metadata.all_fields_info.keys())

    @abstractmethod
    def entity_count(self, name: str) -> int:
        """Return the count of an entity or None if not contained in the archive."""

    @property
    @abstractmethod
    def link_count(self) -> int:
        """Return the count of links."""

    @abstractmethod
    def iter_entity_fields(self,
                           name: str,
                           fields: Optional[Tuple[str, ...]] = None) -> Iterator[Tuple[int, Dict[str, Any]]]:
        """Iterate over entities and yield their pk and database fields."""

    @abstractmethod
    def iter_node_uuids(self) -> Iterator[str]:
        """Iterate over node UUIDs."""

    @abstractmethod
    def iter_group_uuids(self) -> Iterator[Tuple[str, Set[str]]]:
        """Iterate over group UUIDs and the a set of node UUIDs they contain."""

    @abstractmethod
    def iter_link_data(self) -> Iterator[dict]:
        """Iterate over links: {'input': <UUID>, 'output': <UUID>, 'label': <LABEL>, 'type': <TYPE>}"""

    @abstractmethod
    def iter_node_repos(
        self,
        uuids: Iterable[str],
        callback: Callable[[str, Any], None] = null_callback,
    ) -> Iterator[Folder]:
        """Yield temporary folders containing the contents of the repository for each node.

        :param uuids: UUIDs of the nodes over whose repository folders to iterate
        :param callback: a callback to report on the process, ``callback(action, value)``,
            with the following callback signatures:

            - ``callback('init', {'total': <int>, 'description': <str>})``,
               to signal the start of a process, its total iterations and description
            - ``callback('update', <int>)``,
               to signal an update to the process and the number of iterations to progress

        :raises `~aiida.tools.importexport.common.exceptions.CorruptArchive`: If the repository does not exist.
        """

    def node_repository(self, uuid: str) -> Folder:
        """Return a temporary folder with the contents of the repository for a single node.

        :param uuid: The UUID of the node

        :raises `~aiida.tools.importexport.common.exceptions.CorruptArchive`: If the repository does not exist.
        """
        return next(self.iter_node_repos([uuid]))


class ReaderJsonBase(ArchiveReaderAbstract):
    """A reader base for the JSON compressed formats."""

    FILENAME_DATA = 'data.json'
    FILENAME_METADATA = 'metadata.json'
    REPO_FOLDER = NODES_EXPORT_SUBFOLDER

    def __init__(self, filename: str, sandbox_in_repo: bool = False, **kwargs: Any):
        """A reader for JSON compressed archives.

        :param filename: the filename (possibly including the absolute path)
            of the file on which to export.
        :param sandbox_in_repo: Create the temporary uncompressed folder within the aiida repository

        """
        super().__init__(filename, **kwargs)
        self._metadata = None
        self._data = None
        # a temporary folder used to extract the file tree
        self._sandbox: Optional[SandboxFolder] = None
        self._sandbox_in_repo = sandbox_in_repo

    @property
    def file_format_verbose(self) -> str:
        raise NotImplementedError()

    @property
    def compatible_export_version(self) -> str:
        return EXPORT_VERSION

    def __enter__(self):
        super().__enter__()
        self._sandbox = SandboxFolder(self._sandbox_in_repo)
        return self

    def __exit__(
        self, exctype: Optional[Type[BaseException]], excinst: Optional[BaseException], exctb: Optional[TracebackType]
    ):
        self._sandbox.erase()  # type: ignore
        self._sandbox = None
        self._metadata = None
        self._data = None
        super().__exit__(exctype, excinst, exctb)

    def _get_metadata(self):
        """Retrieve the metadata JSON."""
        raise NotImplementedError()

    def _get_data(self):
        """Retrieve the data JSON."""
        raise NotImplementedError()

    def _extract(self, *, path_prefix: str, callback: Callable[[str, Any], None]):
        """Extract repository data to a temporary folder.

        :param path_prefix: Only extract paths starting with this prefix.
        :param callback: a callback to report on the process, ``callback(action, value)``,
            with the following callback signatures:

            - ``callback('init', {'total': <int>, 'description': <str>})``,
               to signal the start of a process, its total iterations and description
            - ``callback('update', <int>)``,
               to signal an update to the process and the number of iterations to progress

        :raises TypeError: if parameter types are not respected
        """
        raise NotImplementedError()

    @property
    def export_version(self) -> str:
        metadata = self._get_metadata()
        if 'export_version' not in metadata:
            raise CorruptArchive('export_version missing from metadata.json')
        return metadata['export_version']

    @property
    def metadata(self) -> ArchiveMetadata:
        metadata = self._get_metadata()
        export_parameters = metadata.get('export_parameters', {})
        output = {
            'export_version': metadata['export_version'],
            'aiida_version': metadata['aiida_version'],
            'all_fields_info': metadata['all_fields_info'],
            'unique_identifiers': metadata['unique_identifiers'],
            'graph_traversal_rules': export_parameters.get('graph_traversal_rules', None),
            'entities_starting_set': export_parameters.get('entities_starting_set', None),
            'include_comments': export_parameters.get('include_comments', None),
            'include_logs': export_parameters.get('include_logs', None),
            'conversion_info': metadata.get('conversion_info', [])
        }
        try:
            return ArchiveMetadata(**output)
        except TypeError as error:
            raise CorruptArchive(f'Metadata invalid: {error}')

    def entity_count(self, name: str) -> int:
        data = self._get_data().get('export_data', {}).get(name, {})
        return len(data)

    @property
    def link_count(self) -> int:
        return len(self._get_data()['links_uuid'])

    def iter_entity_fields(self,
                           name: str,
                           fields: Optional[Tuple[str, ...]] = None) -> Iterator[Tuple[int, Dict[str, Any]]]:
        if name not in self.entity_names:
            raise ValueError(f'Unknown entity name: {name}')
        data = self._get_data()['export_data'].get(name, {})
        if name == NODE_ENTITY_NAME:
            # here we merge in the attributes and extras before yielding
            attributes = self._get_data().get('node_attributes', {})
            extras = self._get_data().get('node_extras', {})
            for pk, all_fields in data.items():
                if pk not in attributes:
                    raise CorruptArchive(f'Unable to find attributes info for Node with Pk={pk}')
                if pk not in extras:
                    raise CorruptArchive(f'Unable to find extra info for Node with Pk={pk}')
                all_fields = {**all_fields, **{'attributes': attributes[pk], 'extras': extras[pk]}}
                if fields is not None:
                    all_fields = {k: v for k, v in all_fields.items() if k in fields}
                yield int(pk), all_fields
        else:
            for pk, all_fields in data.items():
                if fields is not None:
                    all_fields = {k: v for k, v in all_fields.items() if k in fields}
                yield int(pk), all_fields

    def iter_node_uuids(self) -> Iterator[str]:
        for _, fields in self.iter_entity_fields(NODE_ENTITY_NAME, fields=('uuid',)):
            yield fields['uuid']

    def iter_group_uuids(self) -> Iterator[Tuple[str, Set[str]]]:
        group_uuids = self._get_data()['groups_uuid']
        for _, fields in self.iter_entity_fields(GROUP_ENTITY_NAME, fields=('uuid',)):
            key = fields['uuid']
            yield key, set(group_uuids.get(key, set()))

    def iter_link_data(self) -> Iterator[dict]:
        for value in self._get_data()['links_uuid']:
            yield value

    def iter_node_repos(
        self,
        uuids: Iterable[str],
        callback: Callable[[str, Any], None] = null_callback,
    ) -> Iterator[Folder]:
        path_prefixes = [os.path.join(self.REPO_FOLDER, export_shard_uuid(uuid)) for uuid in uuids]

        if not path_prefixes:
            return
        self.assert_within_context()
        assert self._sandbox is not None  # required by mypy

        # unarchive the common folder if it does not exist
        common_prefix = os.path.commonpath(path_prefixes)
        if not self._sandbox.get_subfolder(common_prefix).exists():
            self._extract(path_prefix=common_prefix, callback=callback)

        callback('init', {'total': len(path_prefixes), 'description': 'Iterating node repositories'})
        for uuid, path_prefix in zip(uuids, path_prefixes):
            callback('update', 1)
            subfolder = self._sandbox.get_subfolder(path_prefix)
            if not subfolder.exists():
                raise CorruptArchive(
                    f'Unable to find the repository folder for Node with UUID={uuid} in the exported file'
                )
            yield subfolder


class ReaderJsonZip(ReaderJsonBase):
    """A reader for a JSON zip compressed format."""

    @property
    def file_format_verbose(self) -> str:
        return 'JSON (zip compressed)'

    def _get_metadata(self):
        if self._metadata is None:
            try:
                self._metadata = json.loads(read_file_in_zip(self.filename, self.FILENAME_METADATA))
            except (IOError, FileNotFoundError) as error:
                raise CorruptArchive(str(error))
        return self._metadata

    def _get_data(self):
        if self._data is None:
            try:
                self._data = json.loads(read_file_in_zip(self.filename, self.FILENAME_DATA))
            except (IOError, FileNotFoundError) as error:
                raise CorruptArchive(str(error))
        return self._data

    def _extract(self, *, path_prefix: str, callback: Callable[[str, Any], None] = null_callback):
        self.assert_within_context()
        assert self._sandbox is not None  # required by mypy
        try:
            ZipPath(self.filename, mode='r', allow_zip64=True).joinpath(path_prefix).extract_tree(
                self._sandbox.abspath, callback=callback, cb_descript='Extracting repository files'
            )
        except zipfile.BadZipfile as error:
            raise CorruptArchive(f'The input file cannot be read: {error}')
        except NotADirectoryError as error:
            raise CorruptArchive(f'Unable to find required folder in archive: {error}')


class ReaderJsonTar(ReaderJsonBase):
    """A reader for a JSON tar compressed format."""

    @property
    def file_format_verbose(self) -> str:
        return 'JSON (tar.gz compressed)'

    def _get_metadata(self):
        if self._metadata is None:
            try:
                self._metadata = json.loads(read_file_in_tar(self.filename, self.FILENAME_METADATA))
            except (IOError, FileNotFoundError) as error:
                raise CorruptArchive(str(error))
        return self._metadata

    def _get_data(self):
        if self._data is None:
            try:
                self._data = json.loads(read_file_in_tar(self.filename, self.FILENAME_DATA))
            except (IOError, FileNotFoundError) as error:
                raise CorruptArchive(str(error))
        return self._data

    def _extract(self, *, path_prefix: str, callback: Callable[[str, Any], None] = null_callback):
        self.assert_within_context()
        assert self._sandbox is not None  # required by mypy
        try:
            TarPath(self.filename, mode='r:*').joinpath(path_prefix).extract_tree(
                self._sandbox.abspath,
                allow_dev=False,
                allow_symlink=False,
                callback=callback,
                cb_descript='Extracting repository files'
            )
        except tarfile.ReadError as error:
            raise CorruptArchive(f'The input file cannot be read: {error}')
        except NotADirectoryError as error:
            raise CorruptArchive(f'Unable to find required folder in archive: {error}')


class ReaderJsonFolder(ReaderJsonBase):
    """A reader for a JSON plain folder format."""

    @property
    def file_format_verbose(self) -> str:
        return 'JSON (folder)'

    def _get_metadata(self):
        if self._metadata is None:
            path = Path(self.filename) / self.FILENAME_METADATA
            if not path.exists():
                raise CorruptArchive(f'required file `{self.FILENAME_METADATA}` is not included')
            self._metadata = json.loads(path.read_text(encoding='utf8'))
        return self._metadata

    def _get_data(self):
        if self._data is None:
            path = Path(self.filename) / self.FILENAME_DATA
            if not path.exists():
                raise CorruptArchive(f'required file `{self.FILENAME_DATA}` is not included')
            self._data = json.loads(path.read_text(encoding='utf8'))
        return self._data

    def _extract(self, *, path_prefix: str, callback: Callable[[str, Any], None] = null_callback):
        # pylint: disable=unused-argument
        self.assert_within_context()
        assert self._sandbox is not None  # required by mypy
        # By copying the contents of the source directory, we do not risk to modify the source files accidentally
        # Use path_prefix? or is this quick enough to not worry
        self._sandbox.replace_with_folder(self.filename, overwrite=True)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=import-error,no-name-in-module
"""Utility functions specific to the SqlAlchemy backend."""


def delete_nodes_and_connections_sqla(pks_to_delete):  # pylint: disable=invalid-name
    """
    Delete all nodes corresponding to pks in the input.
    :param pks_to_delete: A list, tuple or set of pks that should be deleted.
    """
    # pylint: disable=no-value-for-parameter
    from aiida.backends.sqlalchemy.models.node import DbNode, DbLink
    from aiida.backends.sqlalchemy.models.group import table_groups_nodes
    from aiida.manage.manager import get_manager

    backend = get_manager().get_backend()

    with backend.transaction() as session:
        # I am first making a statement to delete the membership of these nodes to groups.
        # Since table_groups_nodes is a sqlalchemy.schema.Table, I am using expression language to compile
        # a stmt to be executed by the session. It works, but it's not nice that two different ways are used!
        # Can this be changed?
        stmt = table_groups_nodes.delete().where(table_groups_nodes.c.dbnode_id.in_(list(pks_to_delete)))
        session.execute(stmt)
        # First delete links, then the Nodes, since we are not cascading deletions.
        # Here I delete the links coming out of the nodes marked for deletion.
        session.query(DbLink).filter(DbLink.input_id.in_(list(pks_to_delete))).delete(synchronize_session='fetch')
        # Here I delete the links pointing to the nodes marked for deletion.
        session.query(DbLink).filter(DbLink.output_id.in_(list(pks_to_delete))).delete(synchronize_session='fetch')
        # Now I am deleting the nodes
        session.query(DbNode).filter(DbNode.id.in_(list(pks_to_delete))).delete(synchronize_session='fetch')


def flag_modified(instance, key):
    """Wrapper around `sqlalchemy.orm.attributes.flag_modified` to correctly dereference utils.ModelWrapper

    Since SqlAlchemy 1.2.12 (and maybe earlier but not in 1.0.19) the flag_modified function will check that the
    key is actually present in the instance or it will except. If we pass a model instance, wrapped in the ModelWrapper
    the call will raise an InvalidRequestError. In this function that wraps the flag_modified of SqlAlchemy, we
    derefence the model instance if the passed instance is actually wrapped in the ModelWrapper.
    """
    from sqlalchemy.orm.attributes import flag_modified as flag_modified_sqla
    from aiida.orm.implementation.sqlalchemy.utils import ModelWrapper

    if isinstance(instance, ModelWrapper):
        instance = instance._model  # pylint: disable=protected-access

    flag_modified_sqla(instance, key)


def install_tc(session):
    """
    Install the transitive closure table with SqlAlchemy.
    """
    links_table_name = 'db_dblink'
    links_table_input_field = 'input_id'
    links_table_output_field = 'output_id'
    closure_table_name = 'db_dbpath'
    closure_table_parent_field = 'parent_id'
    closure_table_child_field = 'child_id'

    session.execute(
        get_pg_tc(
            links_table_name, links_table_input_field, links_table_output_field, closure_table_name,
            closure_table_parent_field, closure_table_child_field
        )
    )


def get_pg_tc(
    links_table_name, links_table_input_field, links_table_output_field, closure_table_name, closure_table_parent_field,
    closure_table_child_field
):
    """
    Return the transitive closure table template
    """
    from string import Template

    pg_tc = Template(
        """

DROP TRIGGER IF EXISTS autoupdate_tc ON $links_table_name;
DROP FUNCTION IF EXISTS update_tc();

CREATE OR REPLACE FUNCTION update_tc()
  RETURNS trigger AS
$$BODY$$
DECLARE

    new_id INTEGER;
    old_id INTEGER;
    num_rows INTEGER;

BEGIN

  IF tg_op = 'INSERT' THEN

    IF EXISTS (
      SELECT Id FROM $closure_table_name
      WHERE $closure_table_parent_field = new.$links_table_input_field
         AND $closure_table_child_field = new.$links_table_output_field
         AND depth = 0
         )
    THEN
      RETURN null;
    END IF;

    IF new.$links_table_input_field = new.$links_table_output_field
    OR EXISTS (
      SELECT id FROM $closure_table_name
        WHERE $closure_table_parent_field = new.$links_table_output_field
        AND $closure_table_child_field = new.$links_table_input_field
        )
    THEN
      RETURN null;
    END IF;

    INSERT INTO $closure_table_name (
         $closure_table_parent_field,
         $closure_table_child_field,
         depth)
      VALUES (
         new.$links_table_input_field,
         new.$links_table_output_field,
         0);

    new_id := lastval();

    UPDATE $closure_table_name
      SET entry_edge_id = new_id
        , exit_edge_id = new_id
        , direct_edge_id = new_id
      WHERE id = new_id;


    INSERT INTO $closure_table_name (
      entry_edge_id,
      direct_edge_id,
      exit_edge_id,
      $closure_table_parent_field,
      $closure_table_child_field,
      depth)
      SELECT id
         , new_id
         , new_id
         , $closure_table_parent_field
         , new.$links_table_output_field
         , depth + 1
        FROM $closure_table_name
        WHERE $closure_table_child_field = new.$links_table_input_field;


    INSERT INTO $closure_table_name (
      entry_edge_id,
      direct_edge_id,
      exit_edge_id,
      $closure_table_parent_field,
      $closure_table_child_field,
      depth)
      SELECT new_id
        , new_id
        , id
        , new.$links_table_input_field
        , $closure_table_child_field
        , depth + 1
        FROM $closure_table_name
        WHERE $closure_table_parent_field = new.$links_table_output_field;

    INSERT INTO $closure_table_name (
      entry_edge_id,
      direct_edge_id,
      exit_edge_id,
      $closure_table_parent_field,
      $closure_table_child_field,
      depth)
      SELECT A.id
        , new_id
        , B.id
        , A.$closure_table_parent_field
        , B.$closure_table_child_field
        , A.depth + B.depth + 2
     FROM $closure_table_name A
        CROSS JOIN $closure_table_name B
     WHERE A.$closure_table_child_field = new.$links_table_input_field
       AND B.$closure_table_parent_field = new.$links_table_output_field;

  END IF;

  IF tg_op = 'DELETE' THEN

    IF NOT EXISTS(
        SELECT id FROM $closure_table_name
        WHERE $closure_table_parent_field = old.$links_table_input_field
        AND $closure_table_child_field = old.$links_table_output_field AND
        depth = 0 )
    THEN
        RETURN NULL;
    END IF;

    CREATE TABLE PurgeList (Id int);

    INSERT INTO PurgeList
      SELECT id FROM $closure_table_name
          WHERE $closure_table_parent_field = old.$links_table_input_field
        AND $closure_table_child_field = old.$links_table_output_field AND
        depth = 0;

    WHILE (1 = 1)
    loop

      INSERT INTO PurgeList
        SELECT id FROM $closure_table_name
          WHERE depth > 0
          AND ( entry_edge_id IN ( SELECT Id FROM PurgeList )
          OR direct_edge_id IN ( SELECT Id FROM PurgeList )
          OR exit_edge_id IN ( SELECT Id FROM PurgeList ) )
          AND Id NOT IN (SELECT Id FROM PurgeList );

      GET DIAGNOSTICS num_rows = ROW_COUNT;
      if (num_rows = 0) THEN
        EXIT;
      END IF;
    end loop;

    DELETE FROM $closure_table_name WHERE Id IN ( SELECT Id FROM PurgeList);
    DROP TABLE PurgeList;

  END IF;

  RETURN NULL;

END
$$BODY$$
  LANGUAGE plpgsql VOLATILE
  COST 100;


CREATE TRIGGER autoupdate_tc
  AFTER INSERT OR DELETE OR UPDATE
  ON $links_table_name FOR each ROW
  EXECUTE PROCEDURE update_tc();

"""
    )
    return pg_tc.substitute(
        links_table_name=links_table_name,
        links_table_input_field=links_table_input_field,
        links_table_output_field=links_table_output_field,
        closure_table_name=closure_table_name,
        closure_table_parent_field=closure_table_parent_field,
        closure_table_child_field=closure_table_child_field
    )



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""
.. module::interactive
    :synopsis: Tools and an option class for interactive parameter entry with
    additional features such as help lookup.
"""
import click

from aiida.cmdline.utils import echo
from .conditional import ConditionalOption


class InteractiveOption(ConditionalOption):
    """
    Prompts for input, intercepting certain keyword arguments to replace :mod:`click`'s prompting
    behaviour with a more feature-rich one.

    .. note:: This class has a parameter ``required_fn`` that can be passed to its ``__init__`` (inherited
        from the superclass :py:class:`~aiida.cmdline.params.options.conditional.ConditionalOption`) and a
        ``prompt_fn``.

        - ``required_fn`` is about "is this parameter required" depending on the value of other params.
        - ``prompt_fn`` is about "should I prompt for this value if in interactive mode" and only makes sense
          in this class and not in :py:class:`~aiida.cmdline.params.options.conditional.ConditionalOption`.

        In most usecases, if I have a ``prompt_fn``, then I would like to have also (the same) ``required_fn``.
        The implementation still makes them independent for usecases where they might be different functions
        (e.g. if the variable is anyway not required, but you want to decide whether to prompt for it or not).

    Usage::

        import click

        @click.command()
        @click.option('label', prompt='Label', cls=InteractiveOption)
        def foo(label):
            click.echo(f'Labeling with label: {label}')
    """

    PROMPT_COLOR = 'yellow'
    CHARACTER_PROMPT_HELP = '?'
    CHARACTER_IGNORE_DEFAULT = '!'

    def __init__(self, param_decls=None, switch=None, prompt_fn=None, contextual_default=None, **kwargs):
        """
        :param param_decls: relayed to :class:`click.Option`
        :param switch: sequence of parameter
        :param prompt_fn: callable(ctx) -> True | False, returns True
            if the option should be prompted for in interactive mode.
        :param contextual_default: An optional callback function to get a default which is
            passed the click context

        """
        # intercept prompt kwarg; I need to pop it before calling super
        self._prompt = kwargs.pop('prompt', None)

        # call super class here, after removing `prompt` from the kwargs.
        super().__init__(param_decls=param_decls, **kwargs)

        self.prompt_fn = prompt_fn

        # I check that a prompt was actually defined.
        # I do it after calling super so e.g. 'self.name' is defined
        if not self._prompt:
            raise TypeError(
                f"Interactive options need to have a prompt specified, but '{self.name}' does not have a prompt defined"
            )

        # other kwargs
        self.switch = switch
        self._contextual_default = contextual_default

        # set callback
        self._after_callback = self.callback
        self.callback = self.prompt_callback

        # set control strings that trigger special features from the input prompt
        self._ctrl = {'?': self.ctrl_help}

        # set prompting type
        self.prompt_loop = self.simple_prompt_loop

    @staticmethod
    def is_non_interactive(ctx):
        """Return whether the command is being run non-interactively.

        This is the case if the `non_interactive` parameter in the context is set to `True`.

        :return: boolean, True if being run non-interactively, False otherwise
        """
        return ctx.params.get('non_interactive')

    def get_default(self, ctx):
        """disable :mod:`click` from circumventing prompting when a default value exists"""
        return None

    def _get_default(self, ctx):
        """provides the functionality of :meth:`click.Option.get_default`"""
        if self._contextual_default is not None:
            default = self._contextual_default(ctx)
        else:
            default = super().get_default(ctx)

        try:
            default = self.type.deconvert_default(default)
        except AttributeError:
            pass

        return default

    @staticmethod
    def custom_value_proc(value):
        """Custom value_proc function for the click.prompt which it will call to do value conversion.

        Simply return the value, because we want to take care of value conversion ourselves in the `simple_prompt_loop`.
        If we let `click.prompt` do it, it will raise an exception when the user passes a control character, like the
        question mark, to bring up the help message and the type of the option is not a string, causing conversion to
        fail.
        """
        return value

    def prompt_func(self, ctx):
        """prompt function with args set"""
        return click.prompt(
            click.style(self._prompt, fg=self.PROMPT_COLOR),
            type=self.type,
            value_proc=self.custom_value_proc,
            prompt_suffix=click.style(': ', fg=self.PROMPT_COLOR),
            default=self._get_default(ctx),
            hide_input=self.hide_input,
            confirmation_prompt=self.confirmation_prompt
        )

    def ctrl_help(self):
        """control behaviour when help is requested from the prompt"""
        echo.echo_info(self.format_help_message())

    def format_help_message(self):
        """
        format the message to be displayed for in-prompt help.

        gives a list of possibilities for parameter types that support completion
        """
        msg = self.help or f'Expecting {self.type.name}'
        choices = getattr(self.type, 'complete', lambda x, y: [])(None, '')
        if choices:
            choice_table = []
            msg += '\nSelect one of:\n'
            for choice in choices:
                if isinstance(choice, tuple):
                    choice_table.append('\t{:<12} {}'.format(*choice))
                else:
                    choice_table.append(f'\t{choice:<12}')
            msg += '\n'.join(choice_table)
        return msg

    def full_process_value(self, ctx, value):
        """
        catch errors raised by ConditionalOption in order to adress them in
        the callback
        """
        try:
            value = super().full_process_value(ctx, value)
        except click.MissingParameter:
            pass
        return value

    def safely_convert(self, value, param, ctx):
        """
        convert without raising, instead print a message if fails

        :return: Tuple of ( success (bool), converted value )
        """
        successful = False

        if value is self.CHARACTER_IGNORE_DEFAULT:
            # The ignore default character signifies that the user wants to "not" set the value.
            # Replace value by an empty string for further processing (e.g. if a non-empty value is required).
            value = ''

        try:
            value = self.type.convert(value, param, ctx)
            value = self.callback(ctx, param, value)
            successful = True
        except click.BadParameter as err:
            echo.echo_error(str(err))
            self.ctrl_help()

        return successful, value

    def simple_prompt_loop(self, ctx, param, value):
        """Prompt until successful conversion. dispatch control sequences."""
        if not hasattr(ctx, 'prompt_loop_info_printed'):
            echo.echo_info(f'enter "{self.CHARACTER_PROMPT_HELP}" for help')
            echo.echo_info(f'enter "{self.CHARACTER_IGNORE_DEFAULT}" to ignore the default and set no value')
            ctx.prompt_loop_info_printed = True

        while 1:
            # prompt
            value = self.prompt_func(ctx)
            if value in self._ctrl:
                # dispatch - e.g. show help
                self._ctrl[value]()
                continue

            # try to convert, if unsuccessful continue prompting
            successful, value = self.safely_convert(value, param, ctx)
            if successful:
                return value

    def after_callback(self, ctx, param, value):
        """If a callback was registered on init, call it and return it's value."""
        if self._after_callback:
            try:
                self._after_callback(ctx, param, value)
            except click.BadParameter as exception:
                # If the non-prompt callback raises, we want to only start the prompt loop if we were already in it.
                # For example, if the option was explicitly specified on the command line, but the callback fails, we
                # do not want to start prompting for it, but instead just let the exception bubble-up.
                # However, in this case, the `--non-interactive` flag is not necessarily passed, so we cannot just rely
                # on this value but in addition need to check that we did not already enter the prompt.
                if self.is_non_interactive(ctx) or not hasattr(ctx, 'prompt_loop_info_printed'):
                    raise exception

                echo.echo_error(str(exception))
                self.ctrl_help()
                value = self.prompt_loop(ctx, param, value)

        return value

    def prompt_callback(self, ctx, param, value):
        """decide wether to initiate the prompt_loop or not"""

        # From click.core.Context:
        # if resilient_parsing is enabled then Click will parse without any interactivity or callback invocation.
        # Therefore if this flag is set, we should not do any prompting.
        if ctx.resilient_parsing:
            return None

        # a value was given on the command line: then just go with validation
        if value is not None:
            return self.after_callback(ctx, param, value)

        # The same if the user specified --non-interactive
        if self.is_non_interactive(ctx):

            # Check if it is required
            default = self._get_default(ctx) or self.default

            if default is not None:
                # There is a default
                value = self.type.convert(default, param, ctx)
            else:
                # There is no default.
                # If required
                if self.is_required(ctx):
                    raise click.MissingParameter()
                # In the else case: no default, not required: value is None, it's just  passed to the after_callback
            return self.after_callback(ctx, param, value)

        if self.prompt_fn is None or (self.prompt_fn is not None and self.prompt_fn(ctx)):
            # There is no prompt_fn function, or a prompt_fn function and it says we should ask for the value

            # If we are here, we are in interactive mode and the parameter is not specified
            # We enter the prompt loop
            value = self.prompt_loop(ctx, param, value)
        else:
            # There is a prompt_fn function and returns False (i.e. should not ask for this value
            # We then set the value to None
            value = None

        # And then we call the callback
        return self.after_callback(ctx, param, value)


class TemplateInteractiveOption(InteractiveOption):
    """Sub class of ``InteractiveOption`` that uses template file for input instead of simple inline prompt.

    This is useful for options that need to be able to specify multiline string values.
    """

    def __init__(self, param_decls=None, **kwargs):
        """Define the configuration for the multiline template in the keyword arguments.

        :param template: name of the template to use from the ``aiida.cmdline.templates`` directory.
            Default is the 'multiline.tpl' template.
        :param header: string to put in the header of the template.
        :param footer: string to put in the footer of the template.
        :param extension: file extension to give to the template file.
        """
        self.template = kwargs.pop('template', 'multiline.tpl')
        self.header = kwargs.pop('header', '')
        self.footer = kwargs.pop('footer', '')
        self.extension = kwargs.pop('extension', '')
        super().__init__(param_decls=param_decls, **kwargs)

    def prompt_func(self, ctx):
        """Replace the basic prompt with a method that opens a template file in an editor."""
        from aiida.cmdline.utils.multi_line_input import edit_multiline_template
        kwargs = {'value': self._get_default(ctx) or '', 'header': self.header, 'footer': self.footer}
        return edit_multiline_template(self.template, extension=self.extension, **kwargs)


def opt_prompter(ctx, cmd, givenkwargs, oldvalues=None):
    """
    Prompt interactively for the value of an option of the command with context ``ctx``.

    Used to produce more complex behaviours than can be achieved with InteractiveOption alone.
    """
    if not oldvalues:
        oldvalues = {}
    cmdparams = {i.name: i for i in cmd.params}

    def opt_prompt(opt, prompt, default=None):
        """Prompt interactively for the value of option ``opt``"""
        if not givenkwargs[opt]:
            optobj = cmdparams[opt]
            optobj._prompt = prompt  # pylint: disable=protected-access
            optobj.default = default or oldvalues.get(opt)
            return optobj.prompt_loop(ctx, optobj, givenkwargs[opt])
        return givenkwargs[opt]

    return opt_prompt



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""`verdi data trajectory` command."""

import click

from aiida.cmdline.commands.cmd_data import verdi_data
from aiida.cmdline.commands.cmd_data import cmd_show
from aiida.cmdline.commands.cmd_data.cmd_export import data_export, export_options
from aiida.cmdline.commands.cmd_data.cmd_list import data_list, list_options
from aiida.cmdline.commands.cmd_data.cmd_show import show_options
from aiida.cmdline.params import arguments, options, types
from aiida.cmdline.utils import decorators, echo

LIST_PROJECT_HEADERS = ['Id', 'Label']
EXPORT_FORMATS = ['cif', 'xsf']
VISUALIZATION_FORMATS = ['jmol', 'xcrysden', 'mpl_heatmap', 'mpl_pos']


@verdi_data.group('trajectory')
def trajectory():
    """Manipulate TrajectoryData objects (molecular trajectories)."""


@trajectory.command('list')
@list_options
@decorators.with_dbenv()
def trajectory_list(raw, past_days, groups, all_users):
    """List TrajectoryData objects stored in the database."""
    from aiida.orm import TrajectoryData
    from tabulate import tabulate

    elements = None
    elements_only = False
    formulamode = None
    entry_list = data_list(
        TrajectoryData, LIST_PROJECT_HEADERS, elements, elements_only, formulamode, past_days, groups, all_users
    )

    counter = 0
    struct_list_data = list()
    if not raw:
        struct_list_data.append(LIST_PROJECT_HEADERS)
    for entry in entry_list:
        for i, value in enumerate(entry):
            if isinstance(value, list):
                entry[i] = ','.join(value)
        for i in range(len(entry), len(LIST_PROJECT_HEADERS)):
            entry.append(None)
        counter += 1
    struct_list_data.extend(entry_list)
    if raw:
        echo.echo(tabulate(struct_list_data, tablefmt='plain'))
    else:
        echo.echo(tabulate(struct_list_data, headers='firstrow'))
        echo.echo(f'\nTotal results: {counter}\n')


@trajectory.command('show')
@arguments.DATA(type=types.DataParamType(sub_classes=('aiida.data:array.trajectory',)))
@options.VISUALIZATION_FORMAT(type=click.Choice(VISUALIZATION_FORMATS), default='jmol')
@show_options
@decorators.with_dbenv()
def trajectory_show(data, fmt):
    """Visualize a trajectory."""
    try:
        show_function = getattr(cmd_show, f'_show_{fmt}')
    except AttributeError:
        echo.echo_critical(f'visualization format {fmt} is not supported')

    show_function(fmt, data)


@trajectory.command('export')
@arguments.DATUM(type=types.DataParamType(sub_classes=('aiida.data:array.trajectory',)))
@options.EXPORT_FORMAT(type=click.Choice(EXPORT_FORMATS), default='cif')
@options.TRAJECTORY_INDEX()
@export_options
@decorators.with_dbenv()
def trajectory_export(**kwargs):
    """Export trajectory to file."""
    node = kwargs.pop('datum')
    output = kwargs.pop('output')
    fmt = kwargs.pop('fmt')
    force = kwargs.pop('force')

    kwargs = {k: v for k, v in kwargs.items() if v is not None}

    data_export(node, output, fmt, other_args=kwargs, overwrite=force)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=wildcard-import,undefined-variable
# type: ignore
"""Readers and writers for archive formats, that work independently of a connection to an AiiDA profile."""

from .common import *
from .migrators import *
from .readers import *
from .writers import *

__all__ = (migrators.__all__ + readers.__all__ + writers.__all__ + common.__all__)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=wrong-import-order
"""Setup script for aiida-core package."""
import json
import os

try:
    import fastentrypoints  # pylint: disable=unused-import
except ImportError:
    # This should only occur when building the package, i.e. when
    # executing 'python setup.py sdist' or 'python setup.py bdist_wheel'
    pass
from setuptools import setup, find_packages

if __name__ == '__main__':
    THIS_FOLDER = os.path.split(os.path.abspath(__file__))[0]

    with open(os.path.join(THIS_FOLDER, 'setup.json'), 'r') as info:
        SETUP_JSON = json.load(info)

    EXTRAS_REQUIRE = SETUP_JSON['extras_require']
    EXTRAS_REQUIRE['tests'] = set(EXTRAS_REQUIRE['tests'] + EXTRAS_REQUIRE['rest'] + EXTRAS_REQUIRE['atomic_tools'])
    EXTRAS_REQUIRE['docs'] = set(EXTRAS_REQUIRE['docs'] + EXTRAS_REQUIRE['rest'] + EXTRAS_REQUIRE['atomic_tools'])
    EXTRAS_REQUIRE['all'] = list({item for sublist in EXTRAS_REQUIRE.values() for item in sublist if item != 'bpython'})

    setup(
        packages=find_packages(include=['aiida', 'aiida.*']),
        long_description=open(os.path.join(THIS_FOLDER, 'README.md')).read(),
        long_description_content_type='text/markdown',
        **SETUP_JSON
    )



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=invalid-name,no-member,import-error,no-name-in-module
"""Drop various columns from the `DbUser` model.

These columns were part of the default Django user model

Revision ID: de2eaf6978b4
Revises: 1830c8430131
Create Date: 2019-05-28 11:15:33.242602

"""

from alembic import op
import sqlalchemy as sa
from sqlalchemy.dialects import postgresql

# revision identifiers, used by Alembic.
revision = 'de2eaf6978b4'
down_revision = '1830c8430131'
branch_labels = None
depends_on = None


def upgrade():
    """Migrations for the upgrade."""
    op.drop_column('db_dbuser', 'is_active')
    op.drop_column('db_dbuser', 'is_superuser')
    op.drop_column('db_dbuser', 'is_staff')
    op.drop_column('db_dbuser', 'last_login')
    op.drop_column('db_dbuser', 'password')
    op.drop_column('db_dbuser', 'date_joined')


def downgrade():
    """Migrations for the downgrade."""
    op.add_column(
        'db_dbuser', sa.Column('date_joined', postgresql.TIMESTAMP(timezone=True), autoincrement=False, nullable=True)
    )
    op.add_column('db_dbuser', sa.Column('password', sa.VARCHAR(length=128), autoincrement=False, nullable=True))
    op.add_column(
        'db_dbuser', sa.Column('last_login', postgresql.TIMESTAMP(timezone=True), autoincrement=False, nullable=True)
    )
    op.add_column('db_dbuser', sa.Column('is_staff', sa.BOOLEAN(), autoincrement=False, nullable=True))
    op.add_column('db_dbuser', sa.Column('is_superuser', sa.BOOLEAN(), autoincrement=False, nullable=True))
    op.add_column('db_dbuser', sa.Column('is_active', sa.BOOLEAN(), autoincrement=False, nullable=True))



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Generic `CalcJob` implementation where input file is a parametrized template file."""
import warnings

from aiida.common.warnings import AiidaDeprecationWarning
from aiida.calculations.templatereplacer import TemplatereplacerCalculation  # pylint: disable=unused-import

warnings.warn(  # pylint: disable=no-member
    'The templatereplacer module has moved to aiida.calculations.templatereplacer. '
    'This path will be removed in a future release.', AiidaDeprecationWarning
)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=too-many-nested-blocks,fixme,too-many-arguments,too-many-locals,too-many-branches,too-many-statements
""" SQLAlchemy-specific import of AiiDA entities """
from contextlib import contextmanager
from itertools import chain
from typing import Any, Dict, Iterable, List, Optional, Set, Tuple
import warnings

from sqlalchemy.orm import Session

from aiida.common import json
from aiida.common.links import LinkType
from aiida.common.progress_reporter import get_progress_reporter
from aiida.common.utils import get_object_from_string, validate_uuid
from aiida.common.warnings import AiidaDeprecationWarning
from aiida.orm import QueryBuilder, Node, Group
from aiida.orm.utils.links import link_triple_exists, validate_link

from aiida.tools.importexport.common import exceptions
from aiida.tools.importexport.common.config import DUPL_SUFFIX
from aiida.tools.importexport.common.config import (
    NODE_ENTITY_NAME, GROUP_ENTITY_NAME, COMPUTER_ENTITY_NAME, USER_ENTITY_NAME, LOG_ENTITY_NAME, COMMENT_ENTITY_NAME
)
from aiida.tools.importexport.common.config import (
    entity_names_to_sqla_schema, file_fields_to_model_fields, entity_names_to_entities
)
from aiida.tools.importexport.dbimport.utils import (
    deserialize_field, merge_comment, merge_extras, start_summary, result_summary, IMPORT_LOGGER
)

from aiida.tools.importexport.archive.common import detect_archive_type
from aiida.tools.importexport.archive.readers import ArchiveReaderAbstract, get_reader

from aiida.tools.importexport.dbimport.backends.common import (
    _copy_node_repositories, _make_import_group, _sanitize_extras, MAX_COMPUTERS, MAX_GROUPS
)


def import_data_sqla(
    in_path: str,
    group: Optional[Group] = None,
    ignore_unknown_nodes: bool = False,
    extras_mode_existing: str = 'kcl',
    extras_mode_new: str = 'import',
    comment_mode: str = 'newest',
    silent: Optional[bool] = None,
    **kwargs: Any
):  # pylint: disable=unused-argument
    """Import exported AiiDA archive to the AiiDA database and repository.

    Specific for the SQLAlchemy backend.
    If ``in_path`` is a folder, calls extract_tree; otherwise, tries to detect the compression format
    (zip, tar.gz, tar.bz2, ...) and calls the correct function.

    :param in_path: the path to a file or folder that can be imported in AiiDA.
    :type in_path: str

    :param group: Group wherein all imported Nodes will be placed.
    :type group: :py:class:`~aiida.orm.groups.Group`

    :param extras_mode_existing: 3 letter code that will identify what to do with the extras import.
        The first letter acts on extras that are present in the original node and not present in the imported node.
        Can be either:
        'k' (keep it) or
        'n' (do not keep it).
        The second letter acts on the imported extras that are not present in the original node.
        Can be either:
        'c' (create it) or
        'n' (do not create it).
        The third letter defines what to do in case of a name collision.
        Can be either:
        'l' (leave the old value),
        'u' (update with a new value),
        'd' (delete the extra), or
        'a' (ask what to do if the content is different).
    :type extras_mode_existing: str

    :param extras_mode_new: 'import' to import extras of new nodes or 'none' to ignore them.
    :type extras_mode_new: str

    :param comment_mode: Comment import modes (when same UUIDs are found).
        Can be either:
        'newest' (will keep the Comment with the most recent modification time (mtime)) or
        'overwrite' (will overwrite existing Comments with the ones from the import file).
    :type comment_mode: str

    :param silent: suppress progress bar and summary.
    :type silent: bool

    :return: New and existing Nodes and Links.
    :rtype: dict

    :raises `~aiida.tools.importexport.common.exceptions.ImportValidationError`: if parameters or the contents of
        `metadata.json` or `data.json` can not be validated.
    :raises `~aiida.tools.importexport.common.exceptions.CorruptArchive`: if the provided archive at ``in_path`` is
        corrupted.
    :raises `~aiida.tools.importexport.common.exceptions.IncompatibleArchiveVersionError`: if the provided archive's
        export version is not equal to the export version of AiiDA at the moment of import.
    :raises `~aiida.tools.importexport.common.exceptions.ArchiveImportError`: if there are any internal errors when
        importing.
    :raises `~aiida.tools.importexport.common.exceptions.ImportUniquenessError`: if a new unique entity can not be
        created.
    """
    # Initial check(s)
    if silent is not None:
        warnings.warn(
            'silent keyword is deprecated and will be removed in AiiDA v2.0.0, set the logger level explicitly instead',
            AiidaDeprecationWarning
        )  # pylint: disable=no-member

    if extras_mode_new not in ['import', 'none']:
        raise exceptions.ImportValidationError(
            f"Unknown extras_mode_new value: {extras_mode_new}, should be either 'import' or 'none'"
        )
    reader_cls = get_reader(detect_archive_type(in_path))

    # Initial check(s)
    if group:
        if not isinstance(group, Group):
            raise exceptions.ImportValidationError('group must be a Group entity')
        elif not group.is_stored:
            group.store()

    # The returned dictionary with new and existing nodes and links
    # model_name -> new or existing -> list pk
    ret_dict: Dict[str, dict] = {}

    with reader_cls(in_path) as reader:

        reader.check_version()

        start_summary(in_path, comment_mode, extras_mode_new, extras_mode_existing)

        ##########################################################################
        # CREATE UUID REVERSE TABLES AND CHECK IF I HAVE ALL NODES FOR THE LINKS #
        ##########################################################################
        IMPORT_LOGGER.debug('CHECKING IF NODES FROM LINKS ARE IN DB OR ARCHIVE...')

        linked_nodes = set(chain.from_iterable((l['input'], l['output']) for l in reader.iter_link_data()))
        group_nodes = set(chain.from_iterable((uuids for _, uuids in reader.iter_group_uuids())))

        # Check that UUIDs are valid
        linked_nodes = set(x for x in linked_nodes if validate_uuid(x))
        group_nodes = set(x for x in group_nodes if validate_uuid(x))

        import_nodes_uuid = set(v for v in reader.iter_node_uuids())

        # the combined set of linked_nodes and group_nodes was obtained from looking at all the links
        # the set of import_nodes_uuid was received from the stuff actually referred to in export_data
        unknown_nodes = linked_nodes.union(group_nodes) - import_nodes_uuid

        if unknown_nodes and not ignore_unknown_nodes:
            raise exceptions.DanglingLinkError(
                'The import file refers to {} nodes with unknown UUID, therefore it cannot be imported. Either first '
                'import the unknown nodes, or export also the parents when exporting. The unknown UUIDs are:\n'
                ''.format(len(unknown_nodes)) + '\n'.join('* {}'.format(uuid) for uuid in unknown_nodes)
            )

        ###################################
        # DOUBLE-CHECK MODEL DEPENDENCIES #
        ###################################
        # The entity import order. It is defined by the database model relationships.
        entity_order = [
            USER_ENTITY_NAME, COMPUTER_ENTITY_NAME, NODE_ENTITY_NAME, GROUP_ENTITY_NAME, LOG_ENTITY_NAME,
            COMMENT_ENTITY_NAME
        ]

        for entity_name in reader.entity_names:
            if entity_name not in entity_order:
                raise exceptions.ImportValidationError(f"You are trying to import an unknown model '{entity_name}'!")

        for idx, entity_name in enumerate(entity_order):
            dependencies = []
            # for every field, I checked the dependencies given as value for key requires
            for field in reader.metadata.all_fields_info[entity_name].values():
                try:
                    dependencies.append(field['requires'])
                except KeyError:
                    # (No ForeignKey)
                    pass
            for dependency in dependencies:
                if dependency not in entity_order[:idx]:
                    raise exceptions.ArchiveImportError(
                        f'Entity {entity_name} requires {dependency} but would be loaded first; stopping...'
                    )

        IMPORT_LOGGER.debug('CREATING PK-2-UUID/EMAIL MAPPING...')
        # entity_name -> pk -> unique id
        import_unique_ids_mappings: Dict[str, Dict[int, str]] = {}
        for entity_name, identifier in reader.metadata.unique_identifiers.items():
            import_unique_ids_mappings[entity_name] = {
                int(k): f[identifier] for k, f in reader.iter_entity_fields(entity_name, fields=(identifier,))
            }

        # count total number of entities to import
        number_of_entities: int = sum(reader.entity_count(entity_name) for entity_name in entity_order)
        IMPORT_LOGGER.debug('Importing %s entities', number_of_entities)

        ###########################################
        # IMPORT ALL DATA IN A SINGLE TRANSACTION #
        ###########################################
        with sql_transaction() as session:  # type: Session

            # entity_name -> str(pk) -> fields
            new_entries: Dict[str, Dict[str, dict]] = {}
            existing_entries: Dict[str, Dict[str, dict]] = {}
            # entity_name -> identifier -> pk
            foreign_ids_reverse_mappings: Dict[str, Dict[str, int]] = {}

            IMPORT_LOGGER.debug('ASSESSING IMPORT DATA...')
            for entity_name in entity_order:
                _select_entity_data(
                    entity_name=entity_name,
                    reader=reader,
                    new_entries=new_entries,
                    existing_entries=existing_entries,
                    foreign_ids_reverse_mappings=foreign_ids_reverse_mappings,
                    extras_mode_new=extras_mode_new,
                )

            IMPORT_LOGGER.debug('STORING ENTITIES...')
            for entity_name in entity_order:
                _store_entity_data(
                    reader=reader,
                    entity_name=entity_name,
                    comment_mode=comment_mode,
                    extras_mode_existing=extras_mode_existing,
                    new_entries=new_entries,
                    existing_entries=existing_entries,
                    foreign_ids_reverse_mappings=foreign_ids_reverse_mappings,
                    import_unique_ids_mappings=import_unique_ids_mappings,
                    ret_dict=ret_dict,
                    # batch_size=batch_size,
                    session=session
                )

            # store all pks to add to import group
            pks_for_group: List[int] = [
                foreign_ids_reverse_mappings[NODE_ENTITY_NAME][v['uuid']]
                for entries in [existing_entries, new_entries]
                for v in entries.get(NODE_ENTITY_NAME, {}).values()
            ]

            # now delete the entity data because we no longer need it
            del existing_entries
            del new_entries

            IMPORT_LOGGER.debug('STORING NODE LINKS...')
            _store_node_links(
                reader=reader,
                ignore_unknown_nodes=ignore_unknown_nodes,
                foreign_ids_reverse_mappings=foreign_ids_reverse_mappings,
                ret_dict=ret_dict,
                # batch_size=batch_size,
                session=session
            )

            IMPORT_LOGGER.debug('STORING GROUP ELEMENTS...')
            _add_nodes_to_groups(
                group_count=reader.entity_count(GROUP_ENTITY_NAME),
                group_uuids=reader.iter_group_uuids(),
                foreign_ids_reverse_mappings=foreign_ids_reverse_mappings
            )

        ######################################
        # Put everything in a specific group #
        ######################################
        # Note this is done in a separate transaction
        group = _make_import_group(group=group, node_pks=pks_for_group)

    # Summarize import
    result_summary(ret_dict, getattr(group, 'label', None))

    return ret_dict


@contextmanager
def sql_transaction():
    """A context manager for the transaction."""
    import aiida.backends.sqlalchemy
    session = aiida.backends.sqlalchemy.get_scoped_session()
    try:
        yield session
        IMPORT_LOGGER.debug('COMMITTING EVERYTHING...')
        session.commit()
    except:
        IMPORT_LOGGER.debug('Rolling back')
        session.rollback()
        raise


def _select_entity_data(
    *, entity_name: str, reader: ArchiveReaderAbstract, new_entries: Dict[str, Dict[str, dict]],
    existing_entries: Dict[str, Dict[str, dict]], foreign_ids_reverse_mappings: Dict[str, Dict[str, int]],
    extras_mode_new: str
):
    """Select the data to import by comparing the AiiDA database to the archive contents."""
    entity = entity_names_to_entities[entity_name]

    # entity = entity_names_to_entities[entity_name]
    unique_identifier = reader.metadata.unique_identifiers.get(entity_name, None)

    # Not necessarily all models are present in the archive
    if entity_name not in reader.entity_names:
        return

    existing_entries.setdefault(entity_name, {})
    new_entries.setdefault(entity_name, {})

    if unique_identifier is None:
        new_entries[entity_name] = {str(pk): fields for pk, fields in reader.iter_entity_fields(entity_name)}
        return

    # skip nodes that are already present in the DB
    import_unique_ids = set(
        f[unique_identifier] for _, f in reader.iter_entity_fields(entity_name, fields=(unique_identifier,))
    )

    relevant_db_entries = {}
    if import_unique_ids:
        builder = QueryBuilder()
        builder.append(entity, filters={unique_identifier: {'in': import_unique_ids}}, project='*')

        if builder.count():
            with get_progress_reporter()(
                desc=f'Finding existing entities - {entity_name}', total=builder.count()
            ) as progress:
                for object_ in builder.iterall():
                    progress.update()
                    # Note: UUIDs need to be converted to strings
                    relevant_db_entries.update({str(getattr(object_[0], unique_identifier)): object_[0]})

    foreign_ids_reverse_mappings[entity_name] = {k: v.pk for k, v in relevant_db_entries.items()}

    entity_count = reader.entity_count(entity_name)
    if not entity_count:
        return

    with get_progress_reporter()(desc=f'Reading archived entities - {entity_name}', total=entity_count) as progress:
        imported_comp_names = set()
        for pk, fields in reader.iter_entity_fields(entity_name):
            if entity_name == GROUP_ENTITY_NAME:
                # Check if there is already a group with the same name,
                # and if so, recreate the name
                orig_label = fields['label']
                dupl_counter = 0
                while QueryBuilder().append(entity, filters={'label': {'==': fields['label']}}).count():
                    # Rename the new group
                    fields['label'] = orig_label + DUPL_SUFFIX.format(dupl_counter)
                    dupl_counter += 1
                    if dupl_counter == MAX_GROUPS:
                        raise exceptions.ImportUniquenessError(
                            f'A group of that label ( {orig_label} ) already exists and I could not create a new one'
                        )

            elif entity_name == COMPUTER_ENTITY_NAME:
                # The following is done for compatibility
                # reasons in case the archive file was generated
                # with the Django export method. In Django the
                # metadata and the transport parameters are
                # stored as (unicode) strings of the serialized
                # JSON objects and not as simple serialized
                # JSON objects.
                if isinstance(fields['metadata'], (str, bytes)):
                    fields['metadata'] = json.loads(fields['metadata'])

                # Check if there is already a computer with the
                # same name in the database
                builder = QueryBuilder()
                builder.append(entity, filters={'name': {'==': fields['name']}}, project=['*'], tag='res')
                dupl = builder.count() or fields['name'] in imported_comp_names
                dupl_counter = 0
                orig_name = fields['name']
                while dupl:
                    # Rename the new computer
                    fields['name'] = orig_name + DUPL_SUFFIX.format(dupl_counter)
                    builder = QueryBuilder()
                    builder.append(entity, filters={'name': {'==': fields['name']}}, project=['*'], tag='res')
                    dupl = builder.count() or fields['name'] in imported_comp_names
                    dupl_counter += 1
                    if dupl_counter == MAX_COMPUTERS:
                        raise exceptions.ImportUniquenessError(
                            f'A computer of that name ( {orig_name} ) already exists and I could not create a new one'
                        )

                imported_comp_names.add(fields['name'])

            if fields[unique_identifier] in relevant_db_entries:
                # Already in DB
                existing_entries[entity_name][str(pk)] = fields
            else:
                # To be added
                if entity_name == NODE_ENTITY_NAME:
                    # format extras
                    fields = _sanitize_extras(fields)
                    if extras_mode_new != 'import':
                        fields.pop('extras', None)
                new_entries[entity_name][str(pk)] = fields


def _store_entity_data(
    *, reader: ArchiveReaderAbstract, entity_name: str, comment_mode: str, extras_mode_existing: str,
    new_entries: Dict[str, Dict[str, dict]], existing_entries: Dict[str, Dict[str, dict]],
    foreign_ids_reverse_mappings: Dict[str, Dict[str, int]], import_unique_ids_mappings: Dict[str, Dict[int, str]],
    ret_dict: dict, session: Session
):
    """Store the entity data on the AiiDA profile."""
    from aiida.backends.sqlalchemy.utils import flag_modified
    from aiida.backends.sqlalchemy.models.node import DbNode

    entity = entity_names_to_entities[entity_name]

    fields_info = reader.metadata.all_fields_info.get(entity_name, {})
    unique_identifier = reader.metadata.unique_identifiers.get(entity_name, None)

    pbar_base_str = f'{entity_name}s - '

    # EXISTING ENTRIES
    if existing_entries[entity_name]:

        with get_progress_reporter()(
            total=len(existing_entries[entity_name]), desc=f'{pbar_base_str} existing entries'
        ) as progress:

            for import_entry_pk, entry_data in existing_entries[entity_name].items():

                progress.update()

                unique_id = entry_data[unique_identifier]
                existing_entry_pk = foreign_ids_reverse_mappings[entity_name][unique_id]
                import_data = dict(
                    deserialize_field(
                        k,
                        v,
                        fields_info=fields_info,
                        import_unique_ids_mappings=import_unique_ids_mappings,
                        foreign_ids_reverse_mappings=foreign_ids_reverse_mappings
                    ) for k, v in entry_data.items()
                )
                # TODO COMPARE, AND COMPARE ATTRIBUTES

                if entity_name == COMMENT_ENTITY_NAME:
                    new_entry_uuid = merge_comment(import_data, comment_mode)
                    if new_entry_uuid is not None:
                        entry_data[unique_identifier] = new_entry_uuid
                        new_entries[entity_name][import_entry_pk] = entry_data

                if entity_name not in ret_dict:
                    ret_dict[entity_name] = {'new': [], 'existing': []}
                ret_dict[entity_name]['existing'].append((import_entry_pk, existing_entry_pk))

                # print('  `-> WARNING: NO DUPLICITY CHECK DONE!')
                # CHECK ALSO FILES!

    # Store all objects for this model in a list, and store them all in once at the end.
    objects_to_create = []
    # In the following list we add the objects to be updated
    objects_to_update = []
    # This is needed later to associate the import entry with the new pk
    import_new_entry_pks = {}

    # NEW ENTRIES
    for import_entry_pk, entry_data in new_entries[entity_name].items():
        unique_id = entry_data[unique_identifier]
        import_data = dict(
            deserialize_field(
                k,
                v,
                fields_info=fields_info,
                import_unique_ids_mappings=import_unique_ids_mappings,
                foreign_ids_reverse_mappings=foreign_ids_reverse_mappings
            ) for k, v in entry_data.items()
        )

        # We convert the Django fields to SQLA. Note that some of
        # the Django fields were converted to SQLA compatible
        # fields by the deserialize_field method. This was done
        # for optimization reasons in Django but makes them
        # compatible with the SQLA schema and they don't need any
        # further conversion.
        if entity_name in file_fields_to_model_fields:
            for file_fkey in file_fields_to_model_fields[entity_name]:

                # This is an exception because the DbLog model defines the `_metadata` column instead of the
                # `metadata` column used in the Django model. This is because the SqlAlchemy model base
                # class already has a metadata attribute that cannot be overridden. For consistency, the
                # `DbLog` class however expects the `metadata` keyword in its constructor, so we should
                # ignore the mapping here
                if entity_name == LOG_ENTITY_NAME and file_fkey == 'metadata':
                    continue

                model_fkey = file_fields_to_model_fields[entity_name][file_fkey]
                if model_fkey in import_data:
                    continue
                import_data[model_fkey] = import_data[file_fkey]
                import_data.pop(file_fkey, None)

        db_entity = get_object_from_string(entity_names_to_sqla_schema[entity_name])

        objects_to_create.append(db_entity(**import_data))
        import_new_entry_pks[unique_id] = import_entry_pk

    if entity_name == NODE_ENTITY_NAME:

        # Before storing entries in the DB, I store the files (if these are nodes).
        # Note: only for new entries!
        uuids_to_create = [obj.uuid for obj in objects_to_create]
        _copy_node_repositories(uuids_to_create=uuids_to_create, reader=reader)

        # For the existing nodes that are also in the imported list we also update their extras if necessary
        if existing_entries[entity_name]:

            with get_progress_reporter()(
                total=len(existing_entries[entity_name]), desc='Updating existing node extras'
            ) as progress:

                import_existing_entry_pks = {
                    entry_data[unique_identifier]: import_entry_pk
                    for import_entry_pk, entry_data in existing_entries[entity_name].items()
                }
                for node in session.query(DbNode).filter(DbNode.uuid.in_(import_existing_entry_pks)).all():
                    import_entry_uuid = str(node.uuid)
                    import_entry_pk = import_existing_entry_pks[import_entry_uuid]

                    pbar_node_base_str = f"{pbar_base_str}UUID={import_entry_uuid.split('-')[0]} - "
                    progress.set_description_str(f'{pbar_node_base_str}Extras', refresh=False)
                    progress.update()

                    old_extras = node.extras.copy()
                    extras = existing_entries[entity_name][str(import_entry_pk)].get('extras', {})

                    new_extras = merge_extras(node.extras, extras, extras_mode_existing)

                    if new_extras != old_extras:
                        node.extras = new_extras
                        flag_modified(node, 'extras')
                        objects_to_update.append(node)

    # Store them all in once; However, the PK are not set in this way...
    if objects_to_create:
        session.add_all(objects_to_create)
    if objects_to_update:
        session.add_all(objects_to_update)

    session.flush()

    if not import_new_entry_pks:
        return

    with get_progress_reporter()(total=len(import_new_entry_pks), desc=f'{pbar_base_str} storing new') as progress:

        just_saved = {}

        builder = QueryBuilder()
        builder.append(
            entity,
            filters={unique_identifier: {
                'in': list(import_new_entry_pks.keys())
            }},
            project=[unique_identifier, 'id']
        )

        for entry in builder.iterall():
            progress.update()
            just_saved.update({entry[0]: entry[1]})

        # Now I have the PKs, print the info
        # Moreover, add newly created Nodes to foreign_ids_reverse_mappings
        for unique_id, new_pk in just_saved.items():
            from uuid import UUID
            if isinstance(unique_id, UUID):
                unique_id = str(unique_id)
            import_entry_pk = import_new_entry_pks[unique_id]
            foreign_ids_reverse_mappings[entity_name][unique_id] = new_pk
            if entity_name not in ret_dict:
                ret_dict[entity_name] = {'new': [], 'existing': []}
            ret_dict[entity_name]['new'].append((import_entry_pk, new_pk))

            # TODO prints too many lines
            # IMPORT_LOGGER.debug(f'New {entity_name}: {unique_id} ({import_entry_pk}->{new_pk})')


def _store_node_links(
    *,
    reader: ArchiveReaderAbstract,
    ignore_unknown_nodes: bool,
    foreign_ids_reverse_mappings: Dict[str, Dict[str, int]],
    ret_dict: dict,
    session: Session,
):
    """Store node links to the database."""
    from aiida.backends.sqlalchemy.models.node import DbLink

    link_count = reader.link_count
    if not link_count:
        IMPORT_LOGGER.debug('   (0 new links...)')
        return

    pbar_base_str = 'Links - '
    with get_progress_reporter()(total=link_count, desc=pbar_base_str) as progress_bar:

        for link in reader.iter_link_data():

            progress_bar.set_description_str(f"{pbar_base_str}label={link['label']}", refresh=False)
            progress_bar.update()

            # Check for dangling Links within the, supposed, self-consistent archive
            try:
                in_id = foreign_ids_reverse_mappings[NODE_ENTITY_NAME][link['input']]
                out_id = foreign_ids_reverse_mappings[NODE_ENTITY_NAME][link['output']]
            except KeyError:
                if ignore_unknown_nodes:
                    continue
                raise exceptions.ImportValidationError(
                    'Trying to create a link with one or both unknown nodes, stopping (in_uuid={}, out_uuid={}, '
                    'label={}, type={})'.format(link['input'], link['output'], link['label'], link['type'])
                )

            # Since backend specific Links (DbLink) are not validated upon creation, we will now validate them.
            source = QueryBuilder().append(Node, filters={'id': in_id}, project='*').first()[0]
            target = QueryBuilder().append(Node, filters={'id': out_id}, project='*').first()[0]
            link_type = LinkType(link['type'])

            # Check for existence of a triple link, i.e. unique triple.
            # If it exists, then the link already exists, continue to next link, otherwise, validate link.
            if link_triple_exists(source, target, link_type, link['label']):
                continue

            # Since backend specific Links (DbLink) are not validated upon creation, we will now validate them.
            try:
                validate_link(source, target, link_type, link['label'])
            except ValueError as why:
                raise exceptions.ImportValidationError(f'Error occurred during Link validation: {why}')

            # New link
            session.add(DbLink(input_id=in_id, output_id=out_id, label=link['label'], type=link['type']))
            if 'Link' not in ret_dict:
                ret_dict['Link'] = {'new': []}
            ret_dict['Link']['new'].append((in_id, out_id))

    IMPORT_LOGGER.debug('   (%d new links...)', len(ret_dict.get('Link', {}).get('new', [])))


def _add_nodes_to_groups(
    *, group_count: int, group_uuids: Iterable[Tuple[str, Set[str]]], foreign_ids_reverse_mappings: Dict[str, Dict[str,
                                                                                                                   int]]
):
    """Add nodes to imported groups."""
    if not group_count:
        return

    pbar_base_str = 'Groups - '

    with get_progress_reporter()(total=group_count, desc=pbar_base_str) as progress:
        for groupuuid, groupnodes in group_uuids:
            if not groupnodes:
                progress.update()
                continue
            # TODO: cache these to avoid too many queries
            qb_group = QueryBuilder().append(Group, filters={'uuid': {'==': groupuuid}})
            group_ = qb_group.first()[0]

            progress.set_description_str(f'{pbar_base_str}label={group_.label}', refresh=False)
            progress.update()

            nodes_to_store = [foreign_ids_reverse_mappings[NODE_ENTITY_NAME][node_uuid] for node_uuid in groupnodes]
            qb_nodes = QueryBuilder().append(Node, filters={'id': {'in': nodes_to_store}})
            # Adding nodes to group avoiding the SQLA ORM to increase speed
            nodes_to_add = [n[0].backend_entity for n in qb_nodes.all()]
            group_.backend_entity.add_nodes(nodes_to_add, skip_orm=True)



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""SqlAlchemy implementations for the `Computer` entity and collection."""

from copy import copy

# pylint: disable=import-error,no-name-in-module
from sqlalchemy.exc import SQLAlchemyError
from sqlalchemy.orm.session import make_transient

from aiida.backends.sqlalchemy import get_scoped_session
from aiida.backends.sqlalchemy.models.computer import DbComputer
from aiida.common import exceptions
from aiida.orm.implementation.computers import BackendComputerCollection, BackendComputer

from . import utils
from . import entities


class SqlaComputer(entities.SqlaModelEntity[DbComputer], BackendComputer):
    """SqlAlchemy implementation for `BackendComputer`."""

    # pylint: disable=too-many-public-methods

    MODEL_CLASS = DbComputer

    def __init__(self, backend, **kwargs):
        super().__init__(backend)
        self._dbmodel = utils.ModelWrapper(DbComputer(**kwargs))

    @property
    def uuid(self):
        return str(self._dbmodel.uuid)

    @property
    def pk(self):
        return self._dbmodel.id

    @property
    def id(self):  # pylint: disable=invalid-name
        return self._dbmodel.id

    @property
    def is_stored(self):
        return self._dbmodel.id is not None

    def copy(self):
        """Create an unstored clone of an already stored `Computer`."""
        session = get_scoped_session()

        if not self.is_stored:
            raise exceptions.InvalidOperation('You can copy a computer only after having stored it')

        dbcomputer = copy(self._dbmodel)
        make_transient(dbcomputer)
        session.add(dbcomputer)

        newobject = self.__class__.from_dbmodel(dbcomputer)  # pylint: disable=no-value-for-parameter

        return newobject

    def store(self):
        """Store the `Computer` instance."""
        try:
            self._dbmodel.save()
        except SQLAlchemyError:
            raise ValueError('Integrity error, probably the hostname already exists in the DB')

        return self

    @property
    def name(self):
        return self._dbmodel.name

    @property
    def description(self):
        return self._dbmodel.description

    @property
    def hostname(self):
        return self._dbmodel.hostname

    def get_metadata(self):
        return self._dbmodel._metadata  # pylint: disable=protected-access

    def set_metadata(self, metadata):
        self._dbmodel._metadata = metadata  # pylint: disable=protected-access

    def get_name(self):
        return self._dbmodel.name

    def set_name(self, val):
        self._dbmodel.name = val

    def get_hostname(self):
        return self._dbmodel.hostname

    def set_hostname(self, val):
        self._dbmodel.hostname = val

    def get_description(self):
        return self._dbmodel.description

    def set_description(self, val):
        self._dbmodel.description = val

    def get_scheduler_type(self):
        return self._dbmodel.scheduler_type

    def set_scheduler_type(self, scheduler_type):
        self._dbmodel.scheduler_type = scheduler_type

    def get_transport_type(self):
        return self._dbmodel.transport_type

    def set_transport_type(self, transport_type):
        self._dbmodel.transport_type = transport_type


class SqlaComputerCollection(BackendComputerCollection):
    """Collection of `Computer` instances."""

    ENTITY_CLASS = SqlaComputer

    @staticmethod
    def list_names():
        session = get_scoped_session()
        return session.query(DbComputer.name).all()

    def delete(self, pk):
        try:
            session = get_scoped_session()
            session.query(DbComputer).get(pk).delete()
            session.commit()
        except SQLAlchemyError as exc:
            raise exceptions.InvalidOperation(
                'Unable to delete the requested computer: it is possible that there '
                'is at least one node using this computer (original message: {})'.format(exc)
            )



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
"""Django implementations for the Comment entity and collection."""
# pylint: disable=import-error,no-name-in-module
import contextlib

from datetime import datetime
from django.core.exceptions import ObjectDoesNotExist

from aiida.backends.djsite.db import models
from aiida.common import exceptions, lang

from ..comments import BackendComment, BackendCommentCollection
from .utils import ModelWrapper
from . import entities
from . import users


class DjangoComment(entities.DjangoModelEntity[models.DbComment], BackendComment):
    """Comment implementation for Django."""

    MODEL_CLASS = models.DbComment
    _auto_flush = ('mtime',)

    # pylint: disable=too-many-arguments
    def __init__(self, backend, node, user, content=None, ctime=None, mtime=None):
        """
        Construct a DjangoComment.

        :param node: a Node instance
        :param user: a User instance
        :param content: the comment content
        :param ctime: The creation time as datetime object
        :param mtime: The modification time as datetime object
        :return: a Comment object associated to the given node and user
        """
        super().__init__(backend)
        lang.type_check(user, users.DjangoUser)  # pylint: disable=no-member

        arguments = {
            'dbnode': node.dbmodel,
            'user': user.dbmodel,
            'content': content,
        }

        if ctime:
            lang.type_check(ctime, datetime, f'the given ctime is of type {type(ctime)}')
            arguments['ctime'] = ctime

        if mtime:
            lang.type_check(mtime, datetime, f'the given mtime is of type {type(mtime)}')
            arguments['mtime'] = mtime

        self._dbmodel = ModelWrapper(models.DbComment(**arguments), auto_flush=self._auto_flush)

    def store(self):
        """Can only store if both the node and user are stored as well."""
        from aiida.backends.djsite.db.models import suppress_auto_now

        if self._dbmodel.dbnode.id is None or self._dbmodel.user.id is None:
            raise exceptions.ModificationNotAllowed('The corresponding node and/or user are not stored')

        # `contextlib.suppress` provides empty context and can be replaced with `contextlib.nullcontext` after we drop
        # support for python 3.6
        with suppress_auto_now([(models.DbComment, ['mtime'])]) if self.mtime else contextlib.suppress():
            super().store()

    @property
    def ctime(self):
        return self._dbmodel.ctime

    @property
    def mtime(self):
        return self._dbmodel.mtime

    def set_mtime(self, value):
        self._dbmodel.mtime = value

    @property
    def node(self):
        return self._backend.nodes.from_dbmodel(self._dbmodel.dbnode)

    @property
    def user(self):
        return self._backend.users.from_dbmodel(self._dbmodel.user)

    def set_user(self, value):
        self._dbmodel.user = value

    @property
    def content(self):
        return self._dbmodel.content

    def set_content(self, value):
        self._dbmodel.content = value


class DjangoCommentCollection(BackendCommentCollection):
    """Django implementation for the CommentCollection."""

    ENTITY_CLASS = DjangoComment

    def create(self, node, user, content=None, **kwargs):
        """
        Create a Comment for a given node and user

        :param node: a Node instance
        :param user: a User instance
        :param content: the comment content
        :return: a Comment object associated to the given node and user
        """
        return DjangoComment(self.backend, node, user, content, **kwargs)

    def delete(self, comment_id):
        """
        Remove a Comment from the collection with the given id

        :param comment_id: the id of the comment to delete
        :type comment_id: int

        :raises TypeError: if ``comment_id`` is not an `int`
        :raises `~aiida.common.exceptions.NotExistent`: if Comment with ID ``comment_id`` is not found
        """
        if not isinstance(comment_id, int):
            raise TypeError('comment_id must be an int')

        try:
            models.DbComment.objects.get(id=comment_id).delete()
        except ObjectDoesNotExist:
            raise exceptions.NotExistent(f"Comment with id '{comment_id}' not found")

    def delete_all(self):
        """
        Delete all Comment entries.

        :raises `~aiida.common.exceptions.IntegrityError`: if all Comments could not be deleted
        """
        from django.db import transaction
        try:
            with transaction.atomic():
                models.DbComment.objects.all().delete()
        except Exception as exc:
            raise exceptions.IntegrityError(f'Could not delete all Comments. Full exception: {exc}')

    def delete_many(self, filters):
        """
        Delete Comments based on ``filters``

        :param filters: similar to QueryBuilder filter
        :type filters: dict

        :return: (former) ``PK`` s of deleted Comments
        :rtype: list

        :raises TypeError: if ``filters`` is not a `dict`
        :raises `~aiida.common.exceptions.ValidationError`: if ``filters`` is empty
        """
        from aiida.orm import Comment, QueryBuilder

        # Checks
        if not isinstance(filters, dict):
            raise TypeError('filters must be a dictionary')
        if not filters:
            raise exceptions.ValidationError('filters must not be empty')

        # Apply filter and delete found entities
        builder = QueryBuilder().append(Comment, filters=filters, project='id').all()
        entities_to_delete = [_[0] for _ in builder]
        for entity in entities_to_delete:
            self.delete(entity)

        # Return list of deleted entities' (former) PKs for checking
        return entities_to_delete



# -*- coding: utf-8 -*-
###########################################################################
# Copyright (c), The AiiDA team. All rights reserved.                     #
# This file is part of the AiiDA code.                                    #
#                                                                         #
# The code is hosted on GitHub at https://github.com/aiidateam/aiida-core #
# For further information on the license, see the LICENSE.txt file        #
# For further information please visit http://www.aiida.net               #
###########################################################################
# pylint: disable=wildcard-import,undefined-variable
"""Provides import/export functionalities.

To see history/git blame prior to the move to aiida.tools.importexport,
explore tree: https://github.com/aiidateam/aiida-core/tree/eebef392c81e8b130834a92e1d7abf5e2e30b3ce
Functionality: <tree>/aiida/orm/importexport.py
Tests: <tree>/aiida/backends/tests/test_export_and_import.py
"""
from .archive import *
from .dbexport import *
from .dbimport import *
from .common import *

__all__ = (archive.__all__ + dbexport.__all__ + dbimport.__all__ + common.__all__)
